# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Module for pre-defined widget layouts."""
import ipywidgets as widgets

from .._version import VERSION
from .core import IPyDisplayMixin

__version__ = VERSION
__author__ = "Ian Hellen"


class Progress(IPyDisplayMixin):
    """UI Progress bar."""

    def __init__(self, completed_len: int, visible: bool = True):
        """
        Instantiate new _Progress UI.

        Parameters
        ----------
        completed_len : int
            The expected value that indicates 100% done.
        visible : bool
            If True start the progress UI visible, by default True.

        """
        self._completed = 0
        self._total = completed_len
        self._progress = widgets.IntProgress(
            value=0,
            max=100,
            description="Progress:",
            bar_style="info",
            orientation="horizontal",
        )
        self._done_label = widgets.Label(value="0%")
        if visible:
            self.show()
        else:
            self.hide()
        self.layout = widgets.HBox([self._progress, self._done_label])
        self.display()

    @property
    def value(self) -> int:
        """
        Return the current progress value.

        Returns
        -------
        int
            Progess value

        """
        return self._completed

    @property
    def max(self) -> int:
        """
        Return the current progress maximum value.

        Returns
        -------
        int
            Max value

        """
        return self._total

    def update_progress(self, new_total: int = 0, delta: int = 0):
        """
        Update progress UI by increment or new total.

        Parameters
        ----------
        new_total : int, optional
            New total, by default 0
        delta : int, optional
            Increment to update current total, by default 0

        """
        if new_total:
            self._completed = new_total
        else:
            self._completed += delta
        perc_total = int(100 * self._completed / self._total)
        self._progress.value = perc_total
        self._done_label.value = f"{perc_total}%"

    def show(self):
        """Make the controls visible."""
        self._hide_show("visible")

    def hide(self):
        """Hide the controls."""
        self._hide_show("hidden")

    def _hide_show(self, visibility):
        vis_layout = widgets.Layout(visibility=visibility)
        self._progress.layout = vis_layout
        self._done_label.layout = vis_layout



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""
Initialization sub-package.

This is a sub-package that handles initialization of the package.

- nbinit - notebook initialization
- azure_ml_tools - configuration and helpers for AML workspaces

"""



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""VirusTotal Object browser."""
import pprint
from typing import Dict, Optional

import ipywidgets as widgets
import pandas as pd

from .._version import VERSION
from ..context.vtlookupv3.vtlookupv3 import VTLookupV3, timestamps_to_utcdate
from ..nbwidgets.core import IPyDisplayMixin

__version__ = VERSION
__author__ = "Ian Hellen"


_NOT_FOUND = "Not found"

_BORDER_LAYOUT = widgets.Layout(
    **{
        "width": "90%",
        "border": "solid gray 1px",
        "margin": "1pt",
        "padding": "5pt",
    }
)


# pylint: disable=too-few-public-methods
class VTObjectBrowser(IPyDisplayMixin):
    """VirusTotal object attributes browser."""

    _BASIC_TITLE = "VirusTotal File hash lookup"

    def __init__(self, file_id: Optional[str] = None):
        """
        Initialize the VT Browser.

        Parameters
        ----------
        file_id : Optional[str], optional
            File ID (Hash) of file to be retrieved and displayed, by default None

        """
        self._vt = VTLookupV3()
        self._current_data: pd.DataFrame = pd.DataFrame()

        self.data_sel = widgets.Select(
            description="Attribute", layout=widgets.Layout(height="400px")
        )
        self.data_view = widgets.Textarea(
            description="Value", layout=widgets.Layout(height="400px", width="60%")
        )
        self.data_sel.observe(self._display_attribute, names="value")

        self.txt_file_id = widgets.Text(
            description="Enter file ID (hash)",
            layout=widgets.Layout(width="70%"),
            style={"description_width": "150px"},
            value=file_id or "",
        )

        self.btn_lookup = widgets.Button(description="Lookup")
        self.btn_lookup.on_click(self._lookup_file_id)

        self.html_header = widgets.HTML(
            _summary_html(self._BASIC_TITLE, _extract_summary()), layout=_BORDER_LAYOUT
        )

        self.hb_file_lookup = widgets.HBox(
            [self.txt_file_id, self.btn_lookup], layout=_BORDER_LAYOUT
        )
        self.hb_vt_attribs = widgets.HBox(
            [self.data_sel, self.data_view], layout=_BORDER_LAYOUT
        )
        self.layout = widgets.VBox(
            [self.html_header, self.hb_file_lookup, self.hb_vt_attribs]
        )
        if file_id:
            self.btn_lookup.click()

    def _display_attribute(self, change):
        """Display selected attribute in TextArea."""
        item = change.get("new")
        if item in self._current_data.columns:
            data = self._current_data.iloc[0][item]
            self.data_view.value = pprint.pformat(data)
        else:
            self.data_view.value = ""

    def _lookup_file_id(self, btn):
        """Handle lookup button click."""
        del btn

        self._current_data = self._vt.get_object(self.txt_file_id.value, vt_type="file")
        self.html_header.value = _summary_html(
            self._BASIC_TITLE, _extract_summary(self._current_data)
        )
        if (
            "first_submission_date" not in self._current_data.columns
            or self._current_data.iloc[0]["first_submission_date"] == _NOT_FOUND
        ):
            self.data_sel.options = []
            return
        self._current_data = timestamps_to_utcdate(self._current_data)
        self.data_sel.options = self._current_data.columns


def _extract_summary(data: Optional[pd.DataFrame] = None) -> Dict[str, str]:
    """Return summary of item."""
    def_dict = {"sha256": "", "meaningful_name": "", "names": "", "magic": ""}
    if data is None:
        return def_dict
    if (
        "first_submission_date" not in data.columns
        or data.iloc[0]["first_submission_date"] == _NOT_FOUND
    ):
        def_dict["sha256"] = _NOT_FOUND
        return def_dict
    return data[["sha256", "meaningful_name", "names", "magic"]].iloc[0].to_dict()


def _summary_html(title: str, summary: Dict[str, str]) -> str:
    """Return HTML formatted summary."""
    return f"""
    <h3>{title}</h3>
    <table>
    <tr>
        <td>ID</td><td>{summary.get('sha256')}</td>
    </tr>
    <tr>
        <td>Names</td><td>{summary.get('names')}</td>
    </tr>
    <tr>
        <td>File Type</td><td>{summary.get('magic')}</td>
    </tr>
    </table>
    """



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""
MSTICPY Config modules.

This sub-package contains settings UI management tools
for interactively managing settings in Jupyter notebooks.

It use the ipywidgets package.

"""
from ..lazy_importer import lazy_import

_LAZY_IMPORTS = {
    "msticpy.config.mp_config_control.MpConfigControls",
    "msticpy.config.mp_config_edit.MpConfigEdit",
    "msticpy.config.mp_config_file.MpConfigFile",
}

module, __getattr__, __dir__ = lazy_import(__name__, _LAZY_IMPORTS)



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""
Intsights Provider.

Input can be a single IoC observable or a pandas DataFrame containing
multiple observables. Processing may require a an API key and
processing performance may be limited to a specific number of
requests per minute for the account type that you have.
"""
from __future__ import annotations

import datetime as dt
from dataclasses import dataclass
from typing import Any, ClassVar

from typing_extensions import Self

from ..._version import VERSION
from ...common.utility import export
from ..http_provider import APILookupParams
from .result_severity import ResultSeverity
from .ti_http_provider import HttpTIProvider

__version__ = VERSION
__author__ = "Florian Bracq"


_DEF_HEADERS: dict[str, str] = {
    "Content-Type": "application/json",
    "Accept": "application/json",
}


# pylint: disable=too-few-public-methods
@dataclass
class _IntSightsParams(APILookupParams):
    # override APILookupParams to set common defaults
    def __attrs_post_init__(self: Self) -> None:
        self.auth_str = ["{ApiID}", "{AuthKey}"]
        self.auth_type = "HTTPBasic"


@export
class IntSights(HttpTIProvider):
    """IntSights Lookup."""

    _BASE_URL = "https://api.ti.insight.rapid7.com"

    _QUERIES: ClassVar[dict[str, _IntSightsParams]] = {
        "ipv4": _IntSightsParams(
            path="/public/v3/iocs/ioc-by-value",
            params={"iocValue": "{observable}"},
            headers=_DEF_HEADERS,
        ),
        "ipv6": _IntSightsParams(
            path="/public/v3/iocs/ioc-by-value",
            params={"iocValue": "{observable}"},
            headers=_DEF_HEADERS,
        ),
        "dns": _IntSightsParams(
            path="/public/v3/iocs/ioc-by-value",
            params={"iocValue": "{observable}"},
            headers=_DEF_HEADERS,
        ),
        "url": _IntSightsParams(
            path="/public/v3/iocs/ioc-by-value",
            params={"iocValue": "{observable}"},
            headers=_DEF_HEADERS,
        ),
        "md5_hash": _IntSightsParams(
            path="/public/v3/iocs/ioc-by-value",
            params={"iocValue": "{observable}"},
            headers=_DEF_HEADERS,
        ),
        "sha1_hash": _IntSightsParams(
            path="/public/v3/iocs/ioc-by-value",
            params={"iocValue": "{observable}"},
            headers=_DEF_HEADERS,
        ),
        "sha256_hash": _IntSightsParams(
            path="/public/v3/iocs/ioc-by-value",
            params={"iocValue": "{observable}"},
            headers=_DEF_HEADERS,
        ),
        "email": _IntSightsParams(
            path="/public/v3/iocs/ioc-by-value",
            params={"iocValue": "{observable}"},
            headers=_DEF_HEADERS,
        ),
    }

    _REQUIRED_PARAMS: ClassVar[list[str]] = ["ApiID", "AuthKey"]

    def parse_results(self: Self, response: dict) -> tuple[bool, ResultSeverity, Any]:
        """
        Return the details of the response.

        Parameters
        ----------
        response : Dict
            The returned data response

        Returns
        -------
        Tuple[bool, ResultSeverity, Any]
            bool = positive or negative hit
            ResultSeverity = enumeration of severity
            Object with match details

        """
        if self._failed_response(response) or not isinstance(
            response["RawResult"],
            dict,
        ):
            return False, ResultSeverity.information, "Not found."

        if response["RawResult"].get("whitelisted", False):
            return False, ResultSeverity.information, "Whitelisted."

        sev: str = response["RawResult"].get("severity", "Low")
        result_dict: dict[str, Any] = {
            "threat_actors": response["RawResult"].get("relatedThreatActors", ""),
            "geolocation": response["RawResult"].get("geolocation", None),
            "response_code": response["Status"],
            "tags": response["RawResult"].get("tags", [])
            + response["RawResult"].get("SystemTags", []),
            "malware": response["RawResult"].get("relatedMalware", []),
            "campaigns": response["RawResult"].get("relatedCampaigns", []),
            "score": response["RawResult"].get("score", 0),
            "first_seen": dt.datetime.strptime(
                response["RawResult"].get("firstSeen", None),
                "%Y-%m-%dT%H:%M:%S.%fZ",
            ).replace(tzinfo=dt.timezone.utc),
            "last_seen": dt.datetime.strptime(
                response["RawResult"].get("lastSeen", None),
                "%Y-%m-%dT%H:%M:%S.%fZ",
            ).replace(tzinfo=dt.timezone.utc),
            "last_update": dt.datetime.strptime(
                response["RawResult"].get("lastUpdateDate", None),
                "%Y-%m-%dT%H:%M:%S.%fZ",
            ).replace(tzinfo=dt.timezone.utc),
        }

        severity: ResultSeverity = (
            ResultSeverity.information
            if sev == "Low"
            else (
                ResultSeverity.warning
                if sev == "Medium"
                else ResultSeverity.high if sev == "High" else ResultSeverity.unknown
            )
        )

        return True, severity, result_dict



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Data uploader sub-package."""

try:
    from .loganalytics_uploader import LAUploader  # noqa: F401
    from .splunk_uploader import SplunkUploader  # noqa: F401
except ImportError:
    pass

from ..._version import VERSION

__version__ = VERSION



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""
HTTP Lookup base class.

Input can be a single item or a pandas DataFrame containing
multiple items. Processing may require a an API key and
processing performance may be limited to a specific number of
requests per minute for the account type that you have.

"""
from __future__ import annotations

import traceback
from abc import abstractmethod
from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Any, ClassVar

import httpx
from typing_extensions import Self

from .._version import VERSION
from ..common.exceptions import MsticpyConfigError
from ..common.pkg_config import get_http_timeout
from ..common.utility import mp_ua_header
from .lookup_result import LookupStatus
from .provider_base import Provider

if TYPE_CHECKING:
    import pandas as pd

__version__ = VERSION
__author__ = "Ian Hellen"


@dataclass
class APILookupParams:
    """HTTP Lookup Params definition."""

    path: str = field(default="")
    verb: str = field(default="GET")
    full_url: bool = field(default=False)
    headers: dict[str, str] = field(default_factory=dict)
    params: dict[str, str | float] = field(default_factory=dict)
    data: dict[str, str] = field(default_factory=dict)
    auth_type: str = field(default="")
    auth_str: list[str] = field(default_factory=list)
    sub_type: str = field(default="")


class HttpProvider(Provider):
    """
    HTTP Generic lookup provider base class.

    For subclasses:

    Define Base URL of the service

    .. code:: python

        _BASE_URL = "https://my.api.org/"

    Define query parameters for different item types (keys)

    ..code:: python

        _QUERIES: dict[str, APILookupParams] = {}

    For example:

    .. code:: python

        _QUERIES = {
        # Community API
        "ipv4": APILookupParams(
            path="/v3/community/{observable}",
            headers={"key": "{AuthKey}"},
        ),
        # Enterprise API Quick Lookup
        "ipv4-quick": APILookupParams(
            ...


    Define list of required __init__ params

    .. code:: python

        _REQUIRED_PARAMS: list[str] = []

    For example:

    .. code:: python

        _REQUIRED_PARAMS = ["AuthKey"]

    In __init__

    Be sure to call

    .. code::

        super().__init__(**kwargs)``

    Supply any additional checkers/pre-processors
    with

    .. code: python

        self._preprocessors.add_check(type, check_func)

    See Also
    --------
    PreProcessor
    HttpTIProvider

    """

    # Base URL of the service
    _BASE_URL: ClassVar[str] = ""

    # Define query parameters for different item types (keys)
    _QUERIES: ClassVar[dict[str, Any]] = {}

    # List of required __init__ params
    _REQUIRED_PARAMS: ClassVar[list[str]] = []

    def __init__(
        self: HttpProvider,
        *,
        timeout: int | None = None,
        ApiID: str | None = None,  # noqa: N803
        AuthKey: str | None = None,  # noqa: N803
        Instance: str | None = None,  # noqa: N803
    ) -> None:
        """Initialize the class."""
        super().__init__()
        self._httpx_client = httpx.Client(timeout=get_http_timeout(timeout=timeout))
        self._request_params: dict[str, Any] = {
            "ApiID": None,
            "AuthKey": None,
            "Instance": None,
        }
        if ApiID:
            self._request_params["ApiID"] = ApiID.strip()
        if AuthKey:
            self._request_params["AuthKey"] = AuthKey.strip()
        if Instance:
            self._request_params["Instance"] = Instance.strip()

        missing_params: list[str] = [
            param
            for param in self._REQUIRED_PARAMS
            if param not in self._request_params
        ]

        missing_params = []

        if missing_params:
            param_list: str = ", ".join(f"'{param}'" for param in missing_params)
            error_msg: str = (
                f"Parameter values missing for Provider '{self.__class__.__name__}'. "
                f"Missing parameters are: {param_list}"
            )
            raise MsticpyConfigError(error_msg)

    @abstractmethod
    def lookup_item(
        self: Self,
        item: str,
        item_type: str | None = None,
        query_type: str | None = None,
    ) -> pd.DataFrame:
        """
        Lookup from an item value.

        Parameters
        ----------
        item : str
            item to lookup
        item_type : str, optional
            The Type of the item to lookup, by default None (type will be inferred)
        query_type : str, optional
            Specify the data subtype to be queried, by default None.
            If not specified the default record type for the item_value
            will be returned.

        Returns
        -------
        pd.DataFrame
            The lookup result:
            result - Positive/Negative,
            details - Lookup Details (or status if failure),
            raw_result - Raw Response
            reference - URL of the item

        Raises
        ------
        NotImplementedError
            If attempting to use an HTTP method or authentication
            protocol that is not supported.

        Notes
        -----
        Note: this method uses memoization (lru_cache) to cache results
        for a particular item to try avoid repeated network calls for
        the same item.

        """

    # pylint: enable=duplicate-code
    def _substitute_parms(
        self: Self,
        value: str,
        value_type: str,
        query_type: str | None = None,
    ) -> tuple[str, dict[str, Any]]:
        """
        Create requests parameters collection.

        Parameters
        ----------
        value : str
            The value of the item being queried
        value_type : str, optional
            The value type, by default None
        query_type : str, optional
            Specify the data subtype to be queried, by default None.
            If not specified the default record type for the type
            will be returned.

        Returns
        -------
        Tuple[str, dict[str, Any]]
            HTTP method, dictionary of parameter keys/values

        """
        req_params: dict[str, str] = {"observable": value}
        req_params.update(self._request_params)
        value_key: str = f"{value_type}-{query_type}" if query_type else value_type
        src: APILookupParams | None = self.item_query_defs.get(value_key, None)
        if not src:
            error_msg: str = f"Provider does not support this type {value_key}."
            raise LookupError(error_msg)

        # create a parameter dictionary to pass to requests
        # substitute any parameter value from our req_params dict
        req_dict: dict[str, Any] = {
            "headers": {},
            "url": (
                src.path.format(**req_params)
                if src.full_url
                else (self._BASE_URL + src.path).format(**req_params)
            ),
        }

        if src.headers:
            headers: dict[str, Any] = {
                key: val.format(**req_params) for key, val in src.headers.items()
            }
            req_dict["headers"] = headers
        if "User-Agent" not in req_dict["headers"]:
            req_dict["headers"].update(mp_ua_header())
        if src.params:
            q_params: dict[str, Any] = {
                key: val.format(**req_params) if isinstance(val, str) else val
                for key, val in src.params.items()
            }
            req_dict["params"] = q_params
        if src.data:
            q_data: dict[str, Any] = {
                key: val.format(**req_params) for key, val in src.data.items()
            }
            req_dict["data"] = q_data
        if src.auth_type and src.auth_str:
            auth_strs: tuple = tuple(p.format(**req_params) for p in src.auth_str)
            if src.auth_type == "HTTPBasic":
                req_dict["auth"] = auth_strs
            else:
                error_msg = f"Unknown auth type {src.auth_type}"
                raise NotImplementedError(error_msg)
        return src.verb, req_dict

    @staticmethod
    def _failed_response(response: dict) -> bool:
        """
        Return True if negative response.

        Parameters
        ----------
        response : Dict
            The returned data response

        Returns
        -------
        bool
            True if the response indicated failure.

        """
        return (
            response["Status"] not in (200, LookupStatus.OK.value)
            or not response["RawResult"]
            or not isinstance(response["RawResult"], dict)
        )

    @staticmethod
    def _err_to_results(result: dict, err: Exception) -> None:
        result["Details"] = err.args
        result["RawResult"] = (
            type(err).__name__ + "\n" + str(err) + "\n" + traceback.format_exc()
        )

    @staticmethod
    def _response_message(status_code: int) -> str:
        if status_code == httpx.codes.NOT_FOUND:
            return "Not found."
        if status_code == httpx.codes.UNAUTHORIZED:
            return "Authorization failed. Check account and key details."
        if status_code == httpx.codes.FORBIDDEN:
            return "Request forbidden. Allowed query rate may have been exceeded."
        return httpx.codes.get_reason_phrase(status_code) or "Unknown HTTP status code."



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Widgets sub-package."""

from .._version import VERSION
from ..lazy_importer import lazy_import

__version__ = VERSION
__author__ = "Ian Hellen"

_LAZY_IMPORTS = {
    "msticpy.nbwidgets.core.IPyDisplayMixin",
    "msticpy.nbwidgets.core.RegisteredWidget",
    "msticpy.nbwidgets.get_environment_key.GetEnvironmentKey",
    "msticpy.nbwidgets.get_text.GetText",
    "msticpy.nbwidgets.lookback.Lookback",
    "msticpy.nbwidgets.option_buttons.OptionButtons",
    "msticpy.nbwidgets.progress.Progress",
    "msticpy.nbwidgets.query_time.QueryTime",
    "msticpy.nbwidgets.select_alert.SelectAlert",
    "msticpy.nbwidgets.select_item.SelectItem",
    "msticpy.nbwidgets.select_subset.SelectSubset",
}

module, __getattr__, __dir__ = lazy_import(__name__, _LAZY_IMPORTS)



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Integrates MSTICpy with IPython custom magic commands for retrieval-augmented generation."""

import io
from contextlib import redirect_stdout

from IPython.core.magic import Magics, cell_magic, magics_class
from IPython.display import display_markdown

from .rag_agents import (
    ask_question,
    get_retrieval_assistant_agent,
    get_retrieval_user_proxy_agent,
)


@magics_class
class DocsRagMagic(Magics):
    """Implement a class to provide RAG Magic functionalities for MSTICpy."""

    def __init__(self, shell):
        """Initialize the class with the shell."""
        super().__init__(shell)
        self.assistant_agent = get_retrieval_assistant_agent()
        self.user_proxy_agent = get_retrieval_user_proxy_agent()

    @cell_magic
    def ask(self, _, cell: str):
        """
        Enable the user to ask a question to the RAG agent using a cell magic function.

        It calls the ask_magic method with the cell content as the question.

        Parameters
        ----------
        cell : str
            The content of the cell. This is used as the question to ask the RAG agent.

        Example Usage
        -------------
        To ask a question to the RAG agent, use the cell magic
        command followed by the question in the cell.
        For example:

        %%ask
        Which msticpy module contains the code related to visualizing network graphs?

        """
        question = cell.strip()
        output = io.StringIO()
        with redirect_stdout(output):
            response = ask_question(
                self.assistant_agent,
                self.user_proxy_agent,
                question=question,
            )

        answer = response.summary

        display_markdown(f"\n**Question**: {question}", raw=True)
        display_markdown(f"\n**Answer**: {answer}", raw=True)


# Register the magic class with IPython
def load_ipython_extension(ipython):
    """Register the magic class with IPython."""
    ipython.register_magics(DocsRagMagic)



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Lazy importer for msticpy sub-packages."""
import importlib
from types import ModuleType
from typing import Callable, Iterable, Tuple

from ._version import VERSION

__version__ = VERSION
__author__ = "Ian Hellen"


def lazy_import(
    importer_name: str, import_list: Iterable[str]
) -> Tuple[ModuleType, Callable, Callable]:
    """
    Return the importing module and a callable for lazy importing.

    Parameters
    ----------
    importer_name: str
        The module performing the import to help facilitate resolving
        relative imports.
    import_list:  Iterable[str]
        Iterable of the modules to be potentially imported (absolute
        or relative).
        The basic form is "[mod_path.]import_item". This will import "import_item"
        from the module specified by "mod_path".
        The `as` form of importing is also supported,
        e.g. "pkg.mod.func as spam_func".

    Returns
    -------
    Tuple[module, Callable]
        This function returns a tuple of two items. The first is the importer
        module for easy reference within itself. The second item is a callable to be
        set to `__getattr__` of the calling module.

    Notes
    -----
    Code modified (slightly) from example by Brett Cannon.
    https://snarky.ca/lazy-importing-in-python-3-7/

    """
    module = importlib.import_module(importer_name)
    static_attribs = set(dir(module))
    import_mapping = {}
    for name in import_list:
        importing, _, binding = name.partition(" as ")
        if not binding:
            _, _, binding = importing.rpartition(".")
        import_mapping[binding] = importing

    def __getattr__(name: str):
        """Return the imported module or module member."""
        if name not in import_mapping:
            message = f"module {importer_name!r} has no attribute {name!r}"
            raise AttributeError(message)
        importing = import_mapping[name]
        mod_name, _, attrib_name = importing.rpartition(".")
        if mod_name == importer_name:
            # avoid infinite recursion
            raise AttributeError(
                f"Recursive import of name '[{mod_name}].{name}' from '{importer_name}'."
            )
        # importlib.import_module() implicitly sets submodules on this module as
        # appropriate for direct imports.
        try:
            imported = importlib.import_module(
                mod_name, module.__spec__.parent  # type: ignore
            )
        except ImportError as imp_err:
            message = f"cannot import name '{mod_name}' from '{importer_name}'"
            raise ImportError(message) from imp_err
        mod_attrib = getattr(imported, attrib_name, None)
        setattr(module, name, mod_attrib)
        return mod_attrib

    def __dir__():
        """Return module attribute list combining static and dynamic attribs."""
        return sorted(set(import_mapping).union(static_attribs))

    return module, __getattr__, __dir__



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""TI Providers sub-package."""
from __future__ import annotations

from ..._version import VERSION
from ..preprocess_observable import preprocess_observable  # noqa:F401
from .ti_http_provider import HttpTIProvider  # noqa:F401
from .ti_provider_base import TIProvider  # noqa:F401

__version__ = VERSION

TI_PROVIDERS: dict[str, tuple[str, str]] = {
    "CrowdSec": ("crowdsec", "CrowdSec"),
    "OTX": ("alienvault_otx", "OTX"),
    "AzSTI": ("azure_sent_byoti", "AzSTI"),
    "GreyNoise": ("greynoise", "GreyNoise"),
    "XForce": ("ibm_xforce", "XForce"),
    "IntSights": ("intsights", "IntSights"),
    "OPR": ("open_page_rank", "OPR"),
    "Tor": ("tor_exit_nodes", "Tor"),
    "VirusTotal": ("virustotal", "VirusTotal"),
    "RiskIQ": ("riskiq", "RiskIQ"),
    "Pulsedive": ("pulsedive", "Pulsedive"),
    "AbuseIPDB": ("abuseipdb", "AbuseIPDB"),
    "IPQualityScore": ("ip_quality_score", "IPQualityScore"),
    "BinaryEdge": ("binaryedge", "BinaryEdge"),
}



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Utility classes and functions."""
from __future__ import annotations

import difflib
import inspect
import sys
from enum import Enum
from functools import wraps
from types import ModuleType
from typing import Any, Callable, Iterable, TypeVar, overload

from typing_extensions import Self

from ..._version import VERSION

__version__ = VERSION
__author__ = "Ian Hellen"

T = TypeVar("T")


@overload
def export(obj: type[T]) -> type[T]: ...  # noqa: E704


@overload
def export(obj: Callable) -> Callable: ...  # noqa: E704


def export(obj: type | Callable) -> type | Callable:
    """Decorate function or class to export to __all__."""
    mod: ModuleType = sys.modules[obj.__module__]
    if hasattr(mod, "__all__"):
        all_list: list[str] = getattr(mod, "__all__")
        all_list.append(obj.__name__)
    else:
        all_list = [obj.__name__]
        setattr(mod, "__all__", all_list)
    return obj


@export
def checked_kwargs(legal_args: Iterable[str]):
    """
    Decorate function to check kwargs names against legal arg names.

    Parameters
    ----------
    legal_args : Iterable[str]
        Iterable of possible arguments.

    Raises
    ------
    NameError
        If any of the arguments are not legal. If the an arg is
        a close match to one or more `legal_args`, these are
        returned in the exception.

    Notes
    -----
    The checking is done against the union of legal_args and
    any named arguments of the wrapped function.

    """

    def arg_check_wrapper(func):
        func_args: set[str] = inspect.signature(func).parameters.keys() - {
            "args",
            "kwargs",
        }
        valid_arg_names: set[str] = set(legal_args) | func_args

        @wraps(func)
        def wrapper(*args, **kwargs) -> Callable:
            """Inner argument name checker."""
            name_errs: list[Exception] = []
            for name in kwargs:
                try:
                    check_kwarg(name, valid_arg_names)
                except NameError as err:
                    name_errs.append(err)
            if name_errs:
                raise NameError(name_errs)
            return func(*args, **kwargs)

        return wrapper

    return arg_check_wrapper


@export
def check_kwarg(arg_name: str, legal_args: list[str]) -> None:
    """
    Check argument names against a list.

    Parameters
    ----------
    arg_name : str
        Argument to check
    legal_args : list[str]
        List of possible arguments.

    Raises
    ------
    NameError
        If the argument is not legal. If the `arg_name` is
        a close match to one or more, `legal_args` these are
        returned in the exception.

    """
    if arg_name not in legal_args:
        closest: list[str] = difflib.get_close_matches(arg_name, legal_args)
        msg: str = f"'{arg_name}' is not a recognized argument or attribute. "
        if len(closest) == 1:
            msg += f"Closest match is '{closest[0]}'"
        elif closest:
            match_list: list[str] = [f"'{match}'" for match in closest]
            msg += f"Closest matches are {', '.join(match_list)}"
        else:
            valid_opts: list[str] = [f"'{arg}'" for arg in legal_args]
            msg += f"Valid options are {', '.join(valid_opts)}"
        raise NameError(arg_name, msg)


@export
def check_kwargs(supplied_args: dict[str, Any], legal_args: list[str]) -> None:
    """
    Check all kwargs names against a list.

    Parameters
    ----------
    supplied_args : dict[str, Any]
        Arguments to check
    legal_args : list[str]
        List of possible arguments.

    Raises
    ------
    NameError
        If any of the arguments are not legal. If the an arg is
        a close match to one or more `legal_args`, these are
        returned in the exception.

    """
    name_errs: list[Exception] = []
    for name in supplied_args:
        try:
            check_kwarg(name, legal_args)
        except NameError as err:
            name_errs.append(err)
    if name_errs:
        raise NameError(name_errs)


# Define generic type so enum_parse returns the same type as
# passed in 'enum_class
EnumT = TypeVar("EnumT", bound=Enum)


@export
def enum_parse(enum_cls: type[EnumT], value: str) -> EnumT | None:
    """
    Try to parse a string value to an Enum member.

    Parameters
    ----------
    enum_cls : EnumType
        The Enum type to check against
    value : str
        The enum value to parse

    Returns
    -------
    Optional[EnumType]
        The enumeration value matching `value` or None

    Raises
    ------
    TypeError
        If something other than an Enum subclass is passed.

    """
    if not issubclass(enum_cls, Enum):
        err_msg: str = "Can only be used with classes derived from enum.Enum."
        raise TypeError(err_msg)
    if value in enum_cls.__members__:
        return enum_cls.__members__[value]
    val_lc: str = value.casefold()
    val_map: dict[str, str] = {name.casefold(): name for name in enum_cls.__members__}
    if val_lc in val_map:
        return enum_cls.__members__[val_map[val_lc]]
    return None


@export
class ParseableEnum:
    """Mix-in class for parseable Enum sub-classes."""

    def parse(self: Self, value: str) -> Enum | None:
        """Return enumeration matching (case-insensitive) string value."""
        return enum_parse(enum_cls=self.__class__, value=value)


@export
def arg_to_list(arg: str | list[str], delims: str = ",; ") -> list[str]:
    """
    Convert an optional list/str/str with delims into a list.

    Parameters
    ----------
    arg : Union[str, list[str]]
        A string, delimited string or list
    delims : str, optional
        The default delimiters to use, by default ",; "

    Returns
    -------
    list[str]
        List of string components

    Raises
    ------
    TypeError
        If `arg` is not a string or list

    """
    if isinstance(arg, list):
        return arg
    if isinstance(arg, str):
        for char in delims:
            if char in arg:
                return [item.strip() for item in arg.split(char)]
        return [arg]
    raise TypeError("`arg` must be a string or a list.")


@export
def collapse_dicts(*dicts: dict) -> dict:
    """Merge multiple dictionaries - later dicts have higher precedence."""
    if len(dicts) == 0:
        return {}
    if len(dicts) == 1:
        return dicts[0]
    out_dict: dict = dicts[0]
    for p_dict in dicts[1:]:
        out_dict = _merge_dicts(out_dict, p_dict)
    return out_dict


def _merge_dicts(dict1: dict[Any, Any], dict2: dict[Any, Any]) -> dict:
    """Merge dict2 into dict1."""
    if not dict2:
        return dict1 or {}
    if not dict1:
        return dict2 or {}
    out_dict: dict = {}
    for key in set().union(dict1, dict2):
        if (
            key in dict1
            and isinstance(dict1[key], dict)
            and key in dict2
            and isinstance(dict2[key], dict)
        ):
            d_val: dict = _merge_dicts(dict1[key], dict2[key])
        elif key in dict2:
            d_val = dict2[key]
        else:
            d_val = dict1[key]
        out_dict[key] = d_val
    return out_dict


def singleton(cls: type) -> Callable:
    """Class decorator for singleton classes."""
    instances: dict[type[object], object] = {}

    def get_instance(*args, **kwargs) -> object:
        nonlocal instances
        if cls not in instances:
            instances[cls] = cls(*args, **kwargs)
        return instances[cls]

    return get_instance


@export
class SingletonClass:
    """
    Singleton decorator class.

    Notes
    -----
    Using this decorator on a class enforces the following
    behavior:

    - First instantiation of class will work as normal
    - Subsequent attempts with the same set/values of kwargs
      will just return the original class
    - The class method `current()` will always return the
      last instance of the class.

    """

    def __init__(self: SingletonClass, wrapped_cls: type[Any]) -> None:
        """Instantiate the class wrapper."""
        self.wrapped_cls: type[Any] = wrapped_cls
        self.instance: Self | None = None
        self.__doc__ = wrapped_cls.__doc__

    def __call__(self: Self, *args, **kwargs) -> object:
        """Override the __call__ method for the wrapper class."""
        if self.instance is None:
            self.instance = self.wrapped_cls(*args, **kwargs)
        return self.instance

    def current(self: Self) -> object:
        """Return the current instance of the wrapped class."""
        return self.instance

    def __getattr__(self, name: str) -> Any:
        """Return the attribute `name` from the wrapped class."""
        if hasattr(self.wrapped_cls, name):
            return getattr(self.wrapped_cls, name)
        if self.instance is None:
            return None
        if name == "current":
            return self.instance
        return getattr(self.instance, name)


@export
class SingletonArgsClass(SingletonClass):
    """
    SingletonArgs decorator class.

    Notes
    -----
    Using this decorator on a class enforces the following
    behavior:

    - First instantiation of class will work as normal
    - Subsequent attempts with the same set/values of kwargs
      will just return the original class
    - Instantiation of the class with a different set of kwargs
      will instantiate a new class.
    - The class method `current()` will always return the
      last instance of the class.

    """

    def __init__(self: SingletonArgsClass, wrapped_cls: type[Any]) -> None:
        """Instantiate the class wrapper."""
        super().__init__(wrapped_cls)
        self.kwargs: dict[str, Any] | None = None
        self.args: tuple[Any] | None = None

    def __call__(self, *args, **kwargs) -> object:
        """Override the __call__ method for the wrapper class."""
        if (
            self.instance is None
            or getattr(self.instance, "kwargs", None) != kwargs
            or getattr(self.instance, "args", None) != args
        ):
            self.instance = self.wrapped_cls(*args, **kwargs)
            if self.instance:
                self.instance.kwargs = kwargs
                self.instance.args = args
        return self.instance


@export
class ImportPlaceholder:
    """Placeholder class for optional imports."""

    def __init__(self, name: str, required_pkgs: list[str]) -> None:
        """Initialize class with imported item name and reqd. packages."""
        self.name: str = name
        self.required_pkgs: list[str] = required_pkgs
        self.message: str = (
            f"{self.name} cannot be loaded without the following packages"
            f" installed: {', '.join(self.required_pkgs)}"
        )
        self._mssg_displayed = False

    def _print_req_packages(self) -> None:
        if not self._mssg_displayed:
            print(self.message, "\nPlease install and restart the notebook.")
            self._mssg_displayed = True

    def __getattr__(self, name) -> None:
        """When any attribute is accessed, print requirements."""
        self._print_req_packages()
        raise ImportError(self.name)

    def __call__(self, *args, **kwargs) -> None:
        """If object is called, print requirements."""
        del args, kwargs
        self._print_req_packages()
        raise ImportError(self.name)



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Object container class."""
from typing import Any, Dict, Optional, Type

from .._version import VERSION
from ..common.utility import check_kwarg

__version__ = VERSION
__author__ = "Ian Hellen"


class ObjectContainer:
    """Empty class used to create hierarchical attributes."""

    _subclasses: Dict[str, Type] = {}

    def __len__(self):
        """Return number of items in the attribute collection."""
        return len(self.__dict__)

    def __iter__(self):
        """Return iterator over the attributes."""
        return iter(self.__dict__.items())

    def __getattr__(self, name):
        """Print usable error message if attribute not found."""
        if "." in name:
            try:
                attr = _get_dot_attrib(self, name)
            except KeyError:
                pass
            else:
                return attr
        nm_err: Optional[Exception] = None
        try:
            # check for similar-named attributes in __dict__
            check_kwarg(name, list(self.__dict__.keys()))
        except NameError as err:
            nm_err = err
        if nm_err:
            raise AttributeError(
                f"{self.__class__.__name__} object has no attribute {name}"
            ) from nm_err
        raise AttributeError(
            f"{self.__class__.__name__} object has no attribute {name}"
        )

    def __repr__(self):
        """Return list of attributes."""
        repr_list = []
        for name, obj in self.__dict__.items():
            if isinstance(obj, ObjectContainer):
                repr_list.append(f"{name} (container)")
            elif not name.startswith("_"):
                repr_list.append(f"{name} {type(obj).__name__}")
        return "\n".join(repr_list)

    def __call__(self, *args, **kwargs):
        """Return list of attributes or help."""
        if args or kwargs:
            print("This attribute is a container, not a function.")
            print("Items in this container:")
        print(repr(self))


def _get_dot_attrib(obj, elem_path: str) -> Any:
    """Return attribute at dotted path."""
    path_elems = elem_path.split(".")
    cur_node = obj
    for elem in path_elems:
        cur_node = getattr(cur_node, elem, None)
        if cur_node is None:
            raise KeyError(f"{elem} value of {elem_path} is not a valid path")
    return cur_node


# Descriptors


class SharedProperty:
    """Descriptor to share property between instances of a class."""

    def __init__(self, instance_name: str):
        """
        Initialize the descriptor.

        Parameters
        ----------
        instance_name : str
            Name of the instance attribute to use to store the value.

        """
        self.instance_name = instance_name

    def __get__(self, instance, owner):
        """Return the value of the instance attribute."""
        return getattr(instance, self.instance_name, None)

    def __set__(self, instance, value):
        """Set the value of the instance attribute."""
        setattr(instance, self.instance_name, value)


class FallbackProperty:
    """Descriptor for aliased property with fallback property."""

    def __init__(self, instance_name: str, default_name: str):
        """
        Initialize the descriptor.

        Parameters
        ----------
        instance_name : str
            Name of the instance attribute to use to store the value.
        default_name : str
            Name of the instance attribute to use as a fallback value.

        """
        self.instance_name = instance_name
        self.default_name = default_name

    def __get__(self, instance, owner):
        """Return the value of the instance (or default) attribute."""
        return getattr(
            instance, self.instance_name, getattr(instance, self.default_name, None)
        )

    def __set__(self, instance, value):
        """Set the value of the instance attribute."""
        setattr(instance, self.instance_name, value)


class SplitProperty:
    """Descriptor for property that is stored as two delimited attributes."""

    def __init__(self, inst_left_name: str, inst_right_name: str, split_char: str):
        """
        Initialize the descriptor.

        Parameters
        ----------
        inst_left_name : str
            Name of the instance attribute to use to store the left value.
        inst_right_name : str
            Name of the instance attribute to use to store the right value.
        split_char : str
            Character to use to split the value.

        """
        self.inst_left_name = inst_left_name
        self.inst_right_name = inst_right_name
        self.split_char = split_char

    def __get__(self, instance, owner):
        """Return the value of the instance attribute."""
        left_part = getattr(instance, self.inst_left_name, None)
        right_part = getattr(instance, self.inst_right_name, None)
        if left_part and right_part:
            return f"{left_part}{self.split_char}{right_part}"
        return None

    def __set__(self, instance, value):
        """Set the value of the instance attribute."""
        if value is None:
            return
        if self.split_char not in value:
            setattr(instance, self.inst_left_name, value)
            return
        left, right = value.split(self.split_char, maxsplit=1)
        setattr(instance, self.inst_left_name, left)
        setattr(instance, self.inst_right_name, right)



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Uses the Microsoft Sentinel APIs to interact with Microsoft Sentinel Workspaces."""
from __future__ import annotations

import logging
import warnings
from functools import partial
from typing import TYPE_CHECKING, Any, Callable

from typing_extensions import Self

from ..._version import VERSION
from ...common.exceptions import MsticpyUserConfigError
from ...common.wsconfig import WorkspaceConfig
from .azure_data import get_token
from .sentinel_analytics import SentinelAnalyticsMixin, SentinelHuntingMixin
from .sentinel_dynamic_summary import SentinelDynamicSummaryMixin, SentinelQueryProvider
from .sentinel_incidents import SentinelIncidentsMixin
from .sentinel_search import SentinelSearchlistsMixin
from .sentinel_ti import SentinelTIMixin
from .sentinel_utils import (
    _PATH_MAPPING,
    SentinelInstanceDetails,
    parse_resource_id,
    validate_resource_id,
)
from .sentinel_watchlists import SentinelWatchlistsMixin
from .sentinel_workspaces import SentinelWorkspacesMixin

if TYPE_CHECKING:
    import pandas as pd

__version__ = VERSION
__author__ = "Pete Bryan"

logger: logging.Logger = logging.getLogger(__name__)

_SUB_ID = "subscription_id"
_RES_GRP = "resource_group"
_WS_NAME = "workspace_name"
_RES_ID = "resource_id"


def _create_ws_defaults(
    *,
    subscription_id: str,
    resource_group: str,
    workspace_name: str,
) -> SentinelInstanceDetails:
    """Return default values for Sentinel workspace settings."""
    return SentinelInstanceDetails(
        subscription_id=subscription_id,
        resource_group=resource_group,
        workspace_name=workspace_name,
    )


_LEGACY_PARAM_NAMES: dict[str, str] = {
    "sub_id": _SUB_ID,
    "res_grp": _RES_GRP,
    "ws_name": _WS_NAME,
    "workspace": _WS_NAME,
    "res_id": _RES_ID,
}
_CORE_WS_PARAMETERS: list[str] = [_SUB_ID, _RES_GRP, _WS_NAME]
_WS_PARAMETERS: list[str] = [*_CORE_WS_PARAMETERS, _RES_ID]
_MISSING_PARAMS_ERR: list[str] = [
    "Unable to build a valid resource ID from the parameters provided.",
    "This class requires either a valid Azure resource ID or a combination of",
    "subscription ID, resource group and workspace name.",
    "Please ensure that one of the following is true:",
    (
        "1. You provide a `workspace_name` parameter with corresponding "
        "settings in the configuration file."
    ),
    (
        "2. You provide a valid Azure `resource_id` as parameter to "
        "initialize the class or when calling `connect()`."
    ),
    (
        "3. You provide a valid Azure `subscription_id`, `resource_group` "
        "and `workspace_name` as parameters."
    ),
    "4. Your default workspace (in msticpyconfig) has these settings.",
]


def _map_legacy_param_names(**kwargs) -> dict[str, Any]:
    """
    Map legacy parameter names to current names.

    Parameters
    ----------
    **kwargs : Dict[str, Any]
        The keyword arguments to map.

    Returns
    -------
    Dict[str, Any]
        The mapped keyword arguments with legacy names replaced by current names.

    Notes
    -----
    This is to allow for backward compatibility with older versions of the class.
    These used the following parameter names for workspace settings:
    res_id (resource_id), sub_id (subscription_id), res_grp (resource_group),
    ws_name or workspace (workspace_name).


    """
    return {_LEGACY_PARAM_NAMES.get(key, key): value for key, value in kwargs.items()}


# pylint: disable=too-many-ancestors, too-many-instance-attributes
class MicrosoftSentinel(
    SentinelAnalyticsMixin,
    SentinelHuntingMixin,
    SentinelDynamicSummaryMixin,
    SentinelWatchlistsMixin,
    SentinelSearchlistsMixin,
    SentinelWorkspacesMixin,
    SentinelTIMixin,
    SentinelIncidentsMixin,
):
    """Class for returning key Microsoft Sentinel elements."""

    def __init__(  # pylint:disable=too-many-arguments # noqa:PLR0913
        self: MicrosoftSentinel,
        resource_id: str | None = None,
        *,
        connect: bool = False,
        cloud: str | None = None,
        subscription_id: str | None = None,
        resource_group: str | None = None,
        workspace_name: str | None = None,
        **kwargs,
    ) -> None:
        """
        Initialize connector for Azure APIs.

        Parameters
        ----------
        resource_id : str, optional
            Set the Sentinel workspace resource ID you want to use, if not specified
            defaults will be looked for or details can be passed separately with functions,
            by default None
            `res_id` is an alias for resource_id.
        connect : bool, optional
            Set true if you want to connect to API on initialization, by default False
        cloud : str, optional
            Specify cloud to use, overriding any configuration value.
            Default is to use configuration setting or public cloud if no
            configuration setting is available.
        subscription_id : str, optional
            If not specifying a resource ID the Subscription ID of the Sentinel Workspace
            by default None
            `sub_id` is an alias for subscription_id
        resource_group : str, optional
            If not specifying a resource ID the Resource Group name of the
            Sentinel Workspace, by default None
            `res_grp` is an alias for resource_group
        workspace_name : str, optional
            If not specifying a resource ID, the Workspace name of the
            Sentinel Workspace, by default None
            `ws_name` and `workspace` are aliases for workspace_name
        workspace : str, optional
            If not specifying a resource ID, the Workspace name of the
            Sentinel Workspace, by default None
            `ws_name` and `workspace` are aliases for workspace_name

        Notes
        -----
        There are multiple ways to set the default workspace for the Microsoft Sentinel
        class:
        1. Specify a full Azure resource ID for the workspace in the `resource_id` parameter.
        2. Specify the subscription ID and resource group and workspace name in the
        `subscription_id`, `resource_group` and `workspace` parameters.
        3. Specify only a workspace name in the `workspace` parameter. This will read
        the workspace details from the msticpyconfig configuration file.

        """
        super(SentinelIncidentsMixin, self).__init__(connect=False, cloud=cloud)

        init_kwargs: dict[str, Any] = _map_legacy_param_names(**kwargs)
        if resource_id:
            init_kwargs[_RES_ID] = resource_id
        if subscription_id:
            init_kwargs[_SUB_ID] = subscription_id
        if resource_group:
            init_kwargs[_RES_GRP] = resource_group
        if workspace_name:
            init_kwargs[_WS_NAME] = workspace_name

        self._default_settings: Callable[
            ...,
            SentinelInstanceDetails,
        ] = self._set_ws_defaults(_create_ws_defaults, **init_kwargs)
        self.base_url: str = self.az_cloud_config.resource_manager
        self.sent_data_query: SentinelQueryProvider | None = None
        self.url: str | None = None

        logger.info("Initializing Microsoft Sentinel connector")
        logger.info(
            "Params: Cloud=%s; ResourceId=%s; Workspace=%s, Subscription=%s, ResourceGroup=%s",
            self.cloud,
            self.default_resource_id,
            self.default_workspace_name,
            self.default_subscription_id,
            self.default_resource_group,
        )

        if connect:
            self.connect(**kwargs)

    def connect(  # noqa:PLR0913
        self: Self,
        auth_methods: list[str] | None = None,
        tenant_id: str | None = None,
        *,
        silent: bool = False,
        cloud: str | None = None,
        token: str | None = None,
        **kwargs,
    ) -> None:
        """
        Authenticate with the SDK & API.

        Parameters
        ----------
        auth_methods : List, optional
            list of preferred authentication methods to use, by default None
        tenant_id : str, optional
            Specify cloud tenant to use
        silent : bool, optional
            Set true to prevent output during auth process, by default False
        cloud : str, optional
            What Azure cloud to connect to.
            By default it will attempt to use the cloud setting from config file.
            If this is not set it will default to Azure Public Cloud
        credential: AzureCredential, optional
            Credentials to use for authentication. This will use the credential
            directly and bypass the MSTICPy Azure credential selection process.
        workspace_name: str, optional
            If specified, this will override any default workspace settings
            set during initialization.
            `workspace` is an alias for workspace_name.
        subscription_id: str, optional
            If specified, this will override the subscription ID
            set during initialization.
            `sub_id` is an alias for subscription_id.
        resource_group: str, optional
            If specified, this will override the resource group name
            set during initialization.
            `res_grp` is an alias for resource_group.
        token: str, optional
            If specified, utilize this token to authenticate against Azure.

        Notes
        -----
        You can also supply override the default settings (set during initialization)
        for by supplying either
        1) a full Azure `resource_id` for the workspace or
        2) individual `subscription_id`, `resource_group` and `workspace_name` parameters.
        For this last case, supplying one or more of these will override the default settings
        from initialization for the duration of the authenticated session. E.g. specifying
        a `workspace_name` will override the default workspace name but the subscription ID
        and resource group will remain as set during initialization.

        To revert to the initialization default settings, run `connect()` again without any of
        these parameters.

        See Also
        --------
        msticpy.auth.azure_auth.az_connect : function to authenticate to Azure SDK
        set_default_workspace : method to set the default workspace settings

        """
        connect_kwargs: dict[str, Any] = _map_legacy_param_names(**kwargs)
        if any(connect_kwargs.get(ws_param) for ws_param in _CORE_WS_PARAMETERS):
            try:
                sentinel_instance: SentinelInstanceDetails = SentinelInstanceDetails(
                    subscription_id=connect_kwargs.get(_SUB_ID)
                    or self.default_subscription_id,
                    resource_group=connect_kwargs.get(_RES_GRP)
                    or self.default_resource_group,
                    workspace_name=connect_kwargs.get(_WS_NAME)
                    or self.default_workspace_name,
                )
            except TypeError as err:
                raise MsticpyUserConfigError(
                    *_MISSING_PARAMS_ERR,
                    title="Unable to build valid Sentinel workspace instance",
                ) from err
        else:
            try:
                sentinel_instance = SentinelInstanceDetails.from_resource_id(
                    connect_kwargs.get(_RES_ID) or self.default_resource_id,
                )
            except TypeError as err:
                raise MsticpyUserConfigError(
                    *_MISSING_PARAMS_ERR,
                    title="Unable to build valid Sentinel workspace instance",
                ) from err
        self._create_api_paths_for_workspace(sentinel_instance)

        if cloud is not None and cloud != self.cloud:
            err_msg: str = (
                "Cannot switch to different cloud "
                "and specify the new cloud name using the `cloud` parameter."
            )
            raise MsticpyUserConfigError(
                err_msg,
                title="Cannot switch cloud at connect time",
            )
        logger.info("Using tenant id %s", tenant_id)
        az_connect_kwargs: dict[str, Any] = {
            key: value
            for key, value in connect_kwargs.items()
            if key not in _WS_PARAMETERS
        }
        if tenant_id:
            az_connect_kwargs["tenant_id"] = tenant_id
        self._token = token
        super().connect(auth_methods=auth_methods, silent=silent, **az_connect_kwargs)
        if not self.credentials:
            err_msg = "Could not connect."
            raise ValueError(err_msg)
        if not self._token:
            logger.info("Getting token for %s", tenant_id)
            self._token = get_token(
                self.credentials,
                tenant_id=tenant_id,
                cloud=self.cloud,
            )

    def _create_api_paths_for_workspace(
        self,
        sentinel_instance: SentinelInstanceDetails,
    ) -> None:
        """Save configuration and build API URLs for workspace."""
        try:
            validate_resource_id(sentinel_instance.resource_id)
        except MsticpyUserConfigError as err:
            logger.exception("Error validating resource ID")
            raise MsticpyUserConfigError(
                *_MISSING_PARAMS_ERR,
                title="Unable to build valid resource ID",
            ) from err
        self.url = self._build_sentinel_api_root(
            sentinel_instance=sentinel_instance,
            base_url=self.base_url,
        )

        self.sent_urls = {
            name: f"{self.url}{mapping}" for name, mapping in _PATH_MAPPING.items()
        }
        logger.info("API URLs set to %s", self.sent_urls)

    def set_default_subscription(self: Self, _: str) -> None:
        """Set the default subscription to use to `subscription_id`."""
        err_msg: str = (
            "This method is deprecated. Use `set_default_workspace` instead "
            "or set the subscription ID during initialization."
        )
        raise NotImplementedError(err_msg)

    def set_default_workspace(
        self,
        workspace: str | None = None,
        resource_id: str | None = None,
        **kwargs,
    ) -> None:
        """
        Set the default workspace from workspace name or resource id.

        Parameters
        ----------
        workspace : Optional[str], optional
            Name of the workspace, by default None.
        resource_id: Optional[str], optional
            Azure resource ID for the workspace, by default None.

        Notes
        -----
        If no workspace is specified, the workspace details will be the default
        workspace read from the msticpyconfig configuration file.
        After changing the default workspace, you will need to call `connect()` to
        authenticate with the new workspace.

        """
        adjust_kwargs: dict[str, Any] = _map_legacy_param_names(**kwargs)
        if workspace:
            adjust_kwargs[_WS_NAME] = workspace
        if resource_id:
            adjust_kwargs[_RES_ID] = resource_id
        if _SUB_ID in adjust_kwargs:
            warnings.warn(
                "Setting the workspace from the `subscription_id` parameter "
                "no longer supported. Please use the `workspace` parameter "
                "instead or set the workspace or "
                "Azure resource ID during initialization.",
                stacklevel=1,
            )
        workspace = adjust_kwargs.get(_WS_NAME, workspace)

        if resource_id:
            self._default_settings = self._set_ws_defaults(
                _create_ws_defaults,
                resource_id=resource_id,
            )
            logger.info(
                "Workspace settings set from resource ID %s",
                resource_id,
            )
            return
        # if workspace is one that we have configuration for, get the details from there.
        if workspace in WorkspaceConfig.list_workspaces():
            logger.info("Workspace %s found in settings", workspace)
            workspace_config = WorkspaceConfig(workspace)
        else:
            workspace_config = WorkspaceConfig()
        # set these as defaults
        self._default_settings = self._set_ws_defaults(
            _create_ws_defaults,
            subscription_id=workspace_config.get(WorkspaceConfig.CONF_SUB_ID_KEY),
            resource_group=workspace_config.get(WorkspaceConfig.CONF_RES_GROUP_KEY),
            workspace_name=workspace_config.get(WorkspaceConfig.CONF_WS_NAME_KEY),
        )

    @property
    def default_workspace_settings(self: Self) -> dict[str, Any]:
        """Return current default workspace settings."""
        return WorkspaceConfig.from_settings(
            {
                WorkspaceConfig.CONF_SUB_ID: self.default_subscription_id,
                WorkspaceConfig.CONF_RES_GROUP: self.default_resource_group,
                WorkspaceConfig.CONF_WS_NAME: self.default_workspace_name,
            },
        ).mp_settings

    @property
    def default_subscription_id(self: Self) -> str:
        """Return the default subscription ID."""
        return self._default_settings().subscription_id

    @property
    def default_resource_group(self: Self) -> str:
        """Return the default resource group."""
        return self._default_settings().resource_group

    @property
    def default_workspace_name(self: Self) -> str:
        """Return the default workspace Name."""
        return self._default_settings().workspace_name

    @property
    def default_resource_id(self: Self) -> str:
        """Return the default resource ID."""
        return self._default_settings().resource_id

    def list_data_connectors(self: Self) -> pd.DataFrame:
        """
        List deployed data connectors.

        Returns
        -------
        pd.DataFrame
            A DataFrame containing the deployed data connectors

        Raises
        ------
        CloudError
            If a valid result is not returned.

        """
        return self._list_items(item_type="data_connectors")

    def _set_ws_defaults(
        self,
        create_defaults_func: Callable[..., SentinelInstanceDetails],
        **kwargs,
    ) -> Callable:
        """
        Create a partial function with the defaults set based on the kwargs.

        Parameters
        ----------
        create_defaults_func : Callable
            The function that returns the default values
        **kwargs : Dict[str, Any]
            The keyword arguments for individual workspace settings
            such as subscription_id, resource_group, workspace_name or resource_id

        Returns
        -------
        Callable
            A partial function with the defaults set based on the kwargs

        """
        non_null_kwargs: dict[str, Any] = {
            key: value for key, value in kwargs.items() if value
        }
        workspace_name: str | None = non_null_kwargs.get(_WS_NAME)
        workspace_config: WorkspaceConfig | None = None
        if not any(ws_param in non_null_kwargs for ws_param in _WS_PARAMETERS):
            # if we can't build a resource ID from the parameters, try to get the
            # default workspace settings from the configuration file.
            workspace_config = WorkspaceConfig()

        # If workspace_name is provided, get the config values.
        elif workspace_name and workspace_name in WorkspaceConfig.list_workspaces():
            workspace_config = WorkspaceConfig(workspace=workspace_name)
        if workspace_config:
            config_values: dict[str, Any] = {
                _SUB_ID: workspace_config.get(WorkspaceConfig.CONF_SUB_ID),
                _RES_GRP: workspace_config.get(WorkspaceConfig.CONF_RES_GROUP),
                _WS_NAME: workspace_config.get(WorkspaceConfig.CONF_WS_NAME),
            }
            logger.info("Workspace settings found for %s", workspace_name)
            create_defaults_func = partial(create_defaults_func, **config_values)

        # If any of the individual settings args are provided, set them as defaults
        # Even if they override the config values.
        for arg_name in _CORE_WS_PARAMETERS:
            if arg := non_null_kwargs.get(arg_name):
                create_defaults_func = partial(create_defaults_func, **{arg_name: arg})

        # If an explicit resource ID is provided, parse it and set the defaults
        # This overrides any other settings.
        if resource_id := non_null_kwargs.get(_RES_ID):
            create_defaults_func = partial(
                create_defaults_func,
                **parse_resource_id(resource_id),
            )

        return create_defaults_func


# Alias for old class name
AzureSentinel = MicrosoftSentinel



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""
Open Page Rank Provider.

Input can be a single IoC observable or a pandas DataFrame containing
multiple observables. Processing may require a an API key and
processing performance may be limited to a specific number of
requests per minute for the account type that you have.

"""
from __future__ import annotations

from json import JSONDecodeError
from typing import Any, ClassVar, Iterable

import httpx
import pandas as pd
from typing_extensions import Self

from ..._version import VERSION
from ...common.utility import export
from ..http_provider import APILookupParams
from ..lookup_result import LookupStatus
from ..provider_base import generate_items
from .ti_http_provider import HttpTIProvider
from .ti_provider_base import ResultSeverity

__version__ = VERSION
__author__ = "Ian Hellen"


@export
class OPR(HttpTIProvider):
    """
    Open PageRank Lookup.

    See https://www.domcop.com/openpagerank/what-is-openpagerank
    """

    _BASE_URL = "https://openpagerank.com"

    _QUERIES: ClassVar[dict[str, APILookupParams]] = {
        "dns": APILookupParams(
            path="/api/v1.0/getPageRank",
            params={"domains[0]": "{observable}"},
            headers={"API-OPR": "{AuthKey}"},
        ),
    }

    _REQUIRED_PARAMS: ClassVar[list[str]] = ["AuthKey"]

    def __init__(
        self: OPR,
        *,
        timeout: int | None = None,
        ApiID: str | None = None,  # noqa: N803
        AuthKey: str | None = None,  # noqa: N803
    ) -> None:
        """Initialize a new instance of the class."""
        super().__init__(
            timeout=timeout,
            ApiID=ApiID,
            AuthKey=AuthKey,
        )

        self._provider_name: str = self.__class__.__name__

    async def lookup_iocs_async(
        self: Self,
        data: pd.DataFrame | dict[str, str] | Iterable[str],
        ioc_col: str | None = None,
        ioc_type_col: str | None = None,
        query_type: str | None = None,
    ) -> pd.DataFrame:
        """Call base async wrapper."""
        return await self._lookup_iocs_async_wrapper(
            data,
            ioc_col,
            ioc_type_col,
            query_type,
        )

    def lookup_iocs(
        self: Self,
        data: pd.DataFrame | dict[str, str] | Iterable[str],
        ioc_col: str | None = None,
        ioc_type_col: str | None = None,
        query_type: str | None = None,
    ) -> pd.DataFrame:
        """
        Lookup collection of IoC observables.

        Parameters
        ----------
        data : Union[pd.DataFrame, Dict[str, str], Iterable[str]]
            Data input in one of three formats:
            1. Pandas dataframe (you must supply the column name in
            `obs_col` parameter)
            2. Dict of observable, IoCType
            3. Iterable of observables - IoCTypes will be inferred
        ioc_col : str, optional
            DataFrame column to use for observables, by default None
        ioc_type_col : str, optional
            DataFrame column to use for IoCTypes, by default None
        query_type : str, optional
            Specify the data subtype to be queried, by default None.
            If not specified the default record type for the IoC type
            will be returned.
        provider_name : str, optional
            Name of the provider to query

        Returns
        -------
        pd.DataFrame
            DataFrame of results.

        """
        domain_list: set[str] = set()
        bad_requests: list[pd.Series] = []
        for ioc, ioc_type in generate_items(data, ioc_col, ioc_type_col):
            if not ioc:
                continue
            result: dict[str, Any] = self._check_ioc_type(
                ioc=ioc,
                ioc_type=ioc_type,
                query_subtype=query_type,
            )

            if result["Status"] == LookupStatus.OK.value:
                domain_list.add(result["Ioc"])
            else:
                bad_requests.append(pd.Series(result))

        results: list[pd.Series] = []
        if not domain_list:
            return pd.DataFrame()
        results.extend(
            [pd.Series(elt) for elt in self._lookup_bulk_request(domain_list)],
        )

        all_results: list[pd.Series] = results + bad_requests
        return pd.DataFrame(all_results)

    def parse_results(self: Self, response: dict) -> tuple[bool, ResultSeverity, Any]:
        """
        Return the details of the response.

        Parameters
        ----------
        response : Dict
            The returned data response

        Returns
        -------
        Tuple[bool, ResultSeverity, Any]
            bool = positive or negative hit
            ResultSeverity = enumeration of severity
            Object with match details

        """
        if self._failed_response(response) or not isinstance(
            response["RawResult"],
            dict,
        ):
            return False, ResultSeverity.information, "Not found."

        severity = ResultSeverity.information
        if "response" in response["RawResult"]:
            dom_records: list[dict[str, Any]] = response["RawResult"]["response"]
            dom_record: dict[str, Any] = dom_records[0]
            return self._parse_one_record(dom_record)
        return True, severity, {}

    def _parse_multi_results(self: Self, response: dict) -> Iterable[dict]:
        """Parse details of batch response."""
        if not isinstance(response["RawResult"], dict):
            response.update(
                {
                    "Result": False,
                    "Severity": ResultSeverity.information.value,
                    "Details": "Not found",
                },
            )
            yield response

        elif "response" in response["RawResult"]:
            dom_records: list[dict[str, Any]] = response["RawResult"]["response"]
            for dom_record in dom_records:
                result, sev, details = self._parse_one_record(dom_record)
                domain_name: str = dom_record["domain"]
                yield {
                    "Ioc": domain_name,
                    "IocType": "dns",
                    "Provider": self._provider_name,
                    "Result": result,
                    "Severity": sev.name,
                    "Details": details,
                    "RawResult": dom_record,
                    "Reference": f"{response['Reference']}?domains[0]={domain_name}",
                }

    @staticmethod
    def _parse_one_record(
        dom_record: dict,
    ) -> tuple[bool, ResultSeverity, dict[str, Any]]:
        record_status: int = dom_record.get("status_code", 404)
        severity = ResultSeverity.information
        if record_status == httpx.codes.OK:
            return (
                True,
                severity,
                {
                    "rank": dom_record.get("rank", "0"),
                    "page_rank": dom_record.get("page_rank_decimal", 0),
                    "error": dom_record.get("error", ""),
                },
            )
        if record_status == httpx.codes.NOT_FOUND:
            return (
                True,
                ResultSeverity.warning,
                {
                    "rank": dom_record.get("rank", "0"),
                    "error": dom_record.get("error", ""),
                },
            )
        return False, ResultSeverity.information, {}

    def _lookup_bulk_request(self: Self, ioc_list: Iterable[str]) -> Iterable[dict]:
        ioc_list = list(ioc_list)
        batch_size = 100

        l_len: int = len(ioc_list)
        for step in range(0, l_len, batch_size):
            batch_list: list[str] = ioc_list[step : (step + batch_size)]  # noqa: E203
            yield from self._lookup_batch(batch_list)

    def _lookup_batch(self: Self, ioc_list: list) -> Iterable[dict]:
        # build the query string manually - of the form domains[N]=domN&domains[N+1]...
        qry_elements: list[str] = [
            f"domains[{idx}]={dom}" for idx, dom in zip(range(len(ioc_list)), ioc_list)
        ]

        qry_str: str = "&".join(qry_elements)
        path: str = self._QUERIES["dns"].path
        req_url: str = f"{self._BASE_URL}{path}?{qry_str}"

        try:
            _, req_params = self._substitute_parms("dummy", "dns", None)
            response: httpx.Response = self._httpx_client.get(
                url=req_url,
                headers=req_params["headers"],
            )
            if response.is_success:
                result: dict[str, Any] = {
                    "Ioc": ",".join(ioc_list),
                    "IocType": "dns",
                    "Status": LookupStatus.OK.value,
                    "Reference": f"{self._BASE_URL}{path}",
                    "RawResult": response.json(),
                }

                yield from self._parse_multi_results(result)
            else:
                yield {
                    "Ioc": ",".join(ioc_list),
                    "IocType": "dns",
                    "Status": response.status_code,
                    "Reference": req_url,
                    "RawResult": str(response),
                    "Result": False,
                    "Details": "No response from provider.",
                }
        except (
            LookupError,
            JSONDecodeError,
            NotImplementedError,
            ConnectionError,
        ) as err:
            result_dict: dict[str, str] = {
                "Ioc": ",".join(ioc_list),
                "IocType": "dns",
                "Details": "\n".join(err.args),
            }
            if isinstance(err, LookupError):
                result_dict["Reference"] = req_url
            yield result_dict



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Module for pre-defined widget layouts."""
from datetime import datetime, timedelta
from typing import Optional, Union

import ipywidgets as widgets
from ipywidgets import Layout

from .._version import VERSION
from ..common.timespan import TimeSpan
from ..common.utility import check_kwargs
from .core import (
    IPyDisplayMixin,
    RegisteredWidget,
    TimeUnit,
    default_before_after,
    default_max_buffer,
    parse_time_unit,
)

__version__ = VERSION
__author__ = "Ian Hellen"


# pylint: disable=too-many-instance-attributes
class QueryTime(RegisteredWidget, IPyDisplayMixin):
    """
    QueryTime.

    Composite widget to capture date and time origin
    and set start and end times for queries.

    See Also
    --------
    RegisteredWidget

    """

    _ALLOWED_KWARGS = [
        "origin_time",
        "before",
        "after",
        "start",
        "end",
        "max_before",
        "max_after",
        "label",
        "description",
        "units",
        "auto_display",
        "timespan",
        "register",
        *(RegisteredWidget.ALLOWED_KWARGS),
    ]

    _label_style = {"description_width": "initial"}

    IDS_ATTRIBS = [
        "before",
        "after",
        "_query_start",
        "_query_end",
        "_label",
    ]

    _NB_PARAMS = {"start": "_query_start", "end": "_query_end", "timespan": "timespan"}

    def __init__(
        self,
        **kwargs,
    ):
        """
        Create new instance of QueryTime.

        Parameters
        ----------
        origin_time : datetime, optional
            The origin time (the default is `datetime.utcnow()`)
        description : str, optional
            The description to display
            (the default is 'Select time ({units}) to look back')
            label is an alias for this parameter
        before : int, optional
            The default number of `units` before the `origin_time`
            (the default varies based on the unit)
        after : int, optional
            The default number of `units` after the `origin_time`
            (the default varies based on the unit)
        start : Union[datetime, str]
            Start of query time - alternative to specifying origin,
            before, after
        end : Union[datetime, str]
            End of query time - alternative to specifying origin,
            before, after
        timespan : TimeSpan
            TimeSpan of query time - alternative to specifying origin,
            before, after
        max_before : int, optional
            The largest value for `before` (the default varies based on the unit)
        max_after : int, optional
            The largest value for `after` (the default varies based on the unit)
        units : str, optional
            Time unit (the default is 'hour')
            Permissable values are 'day', 'hour', 'minute', 'second',
            'week'
            These can all be abbreviated down to initial characters
            ('d', 'm', etc.)
        auto_display : bool, optional
            Whether to display on instantiation (the default is False)

        """
        check_kwargs(kwargs, self._ALLOWED_KWARGS)
        self._label = kwargs.pop(
            "description", kwargs.pop("label", "Set query time boundaries")
        )
        self._time_unit = parse_time_unit(kwargs.get("units", "min"))

        self.before = kwargs.pop("before", None)
        self.after = kwargs.pop("after", None)
        self._query_start: Optional[datetime] = None
        self._query_end: Optional[datetime] = None
        self.origin_time = kwargs.pop("origin_time", datetime.utcnow())
        self._get_time_parameters(**kwargs)

        self.max_before = kwargs.pop("max_before", None)
        self.max_after = kwargs.pop("max_after", None)
        self._adjust_max_before_after(self.max_before, self.max_after)

        # Call superclass to register
        ids_params = [
            self.origin_time,
            self.before,
            self.after,
            self.max_before,
            self.max_after,
            self._label,
            self._time_unit,
        ]

        # Create widgets
        self._w_origin_dt = widgets.DatePicker(
            description="Origin Date", disabled=False, value=self.origin_time.date()
        )
        self._w_origin_tm = widgets.Text(
            description="Time (24hr)",
            disabled=False,
            value=str(self.origin_time.time()),
        )

        range_desc = "Time Range"
        # note self.before and self.after are always set to be not None when we
        # reach here (in _get_time_parameters)
        self._w_tm_range = widgets.IntRangeSlider(
            value=(
                -self.before,  # pylint: disable=invalid-unary-operand-type
                self.after,
            ),
            min=-self.max_before,
            max=self.max_after,
            step=1,
            description=range_desc,
            disabled=False,
            continuous_update=True,
            orientation="horizontal",
            readout=True,
            readout_format="d",
            layout=Layout(width="70%"),
            style=self._label_style,
        )
        # pylint: disable=no-member
        self._w_time_unit = widgets.Dropdown(
            options=[
                unit.capitalize()
                for unit, _ in TimeUnit.__members__.items()
                if unit != "Second"
            ],
            value=self._time_unit.name.capitalize(),
            layout=Layout(width="100px"),
        )
        # pylint: enable=no-member

        self._w_start_time_txt = widgets.Text(
            value=self._query_start.isoformat(sep=" "),
            description="Query start time (UTC):",
            layout=Layout(width="50%"),
            style=self._label_style,
        )
        self._w_end_time_txt = widgets.Text(
            value=self._query_end.isoformat(sep=" "),
            description="Query end time (UTC) :  ",
            layout=Layout(width="50%"),
            style=self._label_style,
        )

        super().__init__(id_vals=ids_params, val_attrs=self.IDS_ATTRIBS, **kwargs)
        self._update_ui_controls()
        self._enable_handlers()

        self.layout = self._create_layout()
        if kwargs.pop("auto_display", False):
            self.display()

    def set_time(
        self,
        timespan: Optional[TimeSpan] = None,
        start: Union[datetime, str, None] = None,
        end: Union[datetime, str, None] = None,
    ):
        """
        Change the time attributes.

        Parameters
        ----------
        start : Union[datetime, str]
            Start time - (must specify `end`)
        end : Union[datetime, str]
            Start time - (must also specify `end`)
        timespan : TimeSpan
            TimeSpan for query time.

        """
        if timespan or (start and end):
            self.before = None
            self.after = None
            self._get_time_parameters(
                **{"timespan": timespan, "start": start, "end": end}
            )
        self._update_ui_controls()

    def _create_layout(self):
        return widgets.VBox(
            [
                widgets.HTML(f"<h4>{self._label}</h4>"),
                widgets.HBox([self._w_origin_dt, self._w_origin_tm]),
                widgets.VBox(
                    [
                        widgets.HBox([self._w_tm_range, self._w_time_unit]),
                        self._w_start_time_txt,
                        self._w_end_time_txt,
                    ]
                ),
            ]
        )

    def _get_time_parameters(self, **kwargs):
        """Process different init time parameters from kwargs."""
        timespan: TimeSpan = kwargs.pop("timespan", None)
        start = kwargs.pop("start", None)
        end = kwargs.pop("end", None)
        if timespan:
            self._query_end = self.origin_time = timespan.end
            self._query_start = timespan.start
        elif start and end:
            timespan = TimeSpan(start=start, end=end)
            self._query_start = timespan.start
            self._query_end = self.origin_time = timespan.end
        else:
            self.before = default_before_after(self.before, self._time_unit)
            self.after = default_before_after(self.after, self._time_unit)
            # Calculate time offsets from origin
            self._query_start = self.origin_time - timedelta(
                0, self.before * self._time_unit.value
            )
            self._query_end = self.origin_time + timedelta(
                0, self.after * self._time_unit.value
            )
        if "units" not in kwargs:
            self._infer_time_units()
        if self.after is None:
            self.after = 0
        if self.before is None:
            self.before = int(
                (self._query_end - self._query_start).total_seconds()
                / self._time_unit.value
            )

    # Utility functions
    def _infer_time_units(self):
        """Infer reasonable time unit from current timespan."""
        # If time units not set explicitly, set to something sensible,
        # based on start/end times
        if abs(self.timespan.period.days) > 1:
            self._time_unit = TimeUnit.DAY
        elif abs(self.timespan.period.total_seconds()) > 3600:
            self._time_unit = TimeUnit.HOUR
        else:
            self._time_unit = TimeUnit.MINUTE

    def _adjust_max_before_after(self, max_before, max_after):
        """Adjust the max values so the are always bigger than the defaults."""
        self.max_before = default_max_buffer(
            max_before, self.before or 1, self._time_unit
        )
        self.max_after = default_max_buffer(max_after, self.after or 1, self._time_unit)

    # Widget event handlers
    def _enable_handlers(self):
        # Add change event handlers
        self._w_tm_range.observe(self._time_range_change, names="value")
        self._w_origin_dt.observe(self._update_origin, names="value")
        self._w_origin_tm.observe(self._update_origin, names="value")
        self._w_time_unit.observe(self._change_time_unit, names="value")

    def _disable_handlers(self):
        self._w_tm_range.unobserve_all()
        self._w_origin_dt.unobserve_all()
        self._w_origin_tm.unobserve_all()
        self._w_time_unit.unobserve_all()

    def _change_time_unit(self, change):
        """Handle change event from time unit control."""
        # Reset before/after and max buffers to defaults.
        unit = change["new"]
        self._time_unit = parse_time_unit(unit)
        self.before = default_before_after(default=None, unit=self._time_unit)
        self.after = default_before_after(default=None, unit=self._time_unit)
        self._adjust_max_before_after(max_before=None, max_after=None)
        self._set_time_slider_settings()
        self._update_start_and_end_from_slider()

    def _update_origin(self, change):
        """Handle change events for origin date and time controls."""
        del change
        try:
            tm_value = datetime.strptime(self._w_origin_tm.value, "%H:%M:%S.%f").time()
            self.origin_time = datetime.combine(self._w_origin_dt.value, tm_value)
            self._time_range_change(change=None)
        except (ValueError, TypeError):
            # reset on error
            self._w_origin_dt.value = self.origin_time.date()
            self._w_origin_tm.value = self.origin_time.time().isoformat()

    def _time_range_change(self, change):
        """Handle change event for time slider control."""
        del change
        self._update_start_and_end_from_slider()
        self.before = abs(self._w_tm_range.value[0])
        self.after = abs(self._w_tm_range.value[1])

    def _update_start_and_end_from_slider(self):
        """Update timespan and start and end text controls."""
        self._query_start = self.origin_time + timedelta(
            0, self._w_tm_range.value[0] * self._time_unit.value
        )
        self._query_end = self.origin_time + timedelta(
            0, self._w_tm_range.value[1] * self._time_unit.value
        )
        self._w_start_time_txt.value = self._query_start.isoformat(sep=" ")
        self._w_end_time_txt.value = self._query_end.isoformat(sep=" ")

    # end - event handlers

    # control updates from attributes
    def _update_ui_controls(self):
        """Update UI controls from attributes."""
        self._disable_handlers()
        self._w_start_time_txt.value = self._query_start.isoformat(sep=" ")
        self._w_end_time_txt.value = self._query_end.isoformat(sep=" ")
        self._set_time_slider_settings()
        self._w_origin_dt.value = self.origin_time.date()
        self._w_origin_tm.value = self.origin_time.time().isoformat()
        self._w_time_unit.value = self._time_unit.name.capitalize()
        self._enable_handlers()

    def _set_time_slider_settings(self):
        """Set slider properties based on current before and after values."""
        self._w_tm_range.value = (
            -self.before,  # pylint: disable=invalid-unary-operand-type
            self.after,
        )
        self._w_tm_range.min = -self.max_before
        self._w_tm_range.max = self.max_after

    # end control updates from attributes

    @property
    def start(self):
        """Query start time."""
        return self._query_start

    @property
    def end(self):
        """Query end time."""
        return self._query_end

    @property
    def units(self):
        """Time units used by control."""
        return self._time_unit.name.capitalize()

    @property
    def timespan(self):
        """Return the timespan as a TimeSpan object."""
        return TimeSpan(start=self.start, end=self.end)

    @timespan.setter
    def timespan(self, value: TimeSpan):
        """Set the timespan of the QueryTime widget."""
        self.set_time(timespan=value)

    @property
    def value(self):
        """Return the timespan as a TimeSpan object."""
        return self.timespan

    @value.setter
    def value(self, value: TimeSpan):
        """Set the timespan of the QueryTime widget."""
        self.set_time(timespan=value)



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Module for converting DataFrame to Networkx graph."""
from typing import Callable, Dict, Iterable, Optional, Union

import networkx as nx
import pandas as pd
from typing_extensions import Literal

from .._version import VERSION

__version__ = VERSION
__author__ = "Ian Hellen"

GraphType = Literal["graph", "digraph"]
NodeRole = Literal["source", "target"]
GraphItem = Literal["node", "edge"]


def df_to_networkx(
    data: pd.DataFrame,
    source_col: str,
    target_col: str,
    source_attrs: Optional[Iterable[str]] = None,
    target_attrs: Optional[Iterable[str]] = None,
    edge_attrs: Optional[Iterable[str]] = None,
    graph_type: GraphType = "graph",
):
    """
    Create a networkx graph from a DataFrame.

    Parameters
    ----------
    data : pd.DataFrame
       Input data
    source_col : str
        Column for source nodes.
    target_col : str
        Column for target nodes.
    source_attrs : Optional[List[str]], optional
        Optional list of columns to use as source node attributes, by default None
    target_attrs : Optional[List[str]], optional
        Optional list of columns to use as target node attributes, by default None
    edge_attrs : Optional[List[str]], optional
        Optional list of columns to use as edge node attributes, by default None
    graph_type : str
        "graph" or "digraph" (for nx.DiGraph)

    Returns
    -------
    nx.Graph
        The networkx graph object

    """
    create_as = nx.DiGraph if graph_type == "digraph" else nx.Graph
    _verify_columns(
        data, source_col, target_col, source_attrs, target_attrs, edge_attrs
    )
    # remove any source or target rows that are NaN
    data = data.dropna(axis=0, subset=[source_col, target_col])
    nx_graph = nx.from_pandas_edgelist(
        data,
        source=source_col,
        target=target_col,
        edge_attr=edge_attrs,
        create_using=create_as,
    )

    _set_node_attributes(data, nx_graph, source_col, source_attrs, node_role="source")
    _set_node_attributes(data, nx_graph, target_col, target_attrs, node_role="target")
    return nx_graph


def _set_node_attributes(
    data: pd.DataFrame,
    graph: nx.Graph,
    column: str,
    attrib_cols: Optional[Iterable[str]],
    node_role: NodeRole,
):
    """Set node attributes from column values."""
    all_cols = [column, *attrib_cols] if attrib_cols else [column]
    # Create an 'agg' dictionary to apply to DataFrame
    agg_dict: Dict[str, Union[str, Callable]] = (
        {col: _pd_unique_list for col in attrib_cols} if attrib_cols else {}
    )
    # Add these two items as attributes
    agg_dict.update({"node_role": "first", "node_type": "first"})

    # Group by the column value and apply the agg dictionary
    # to_dict(orient="index") produces a dict like
    # { column_val: { attrib_name: attrib_val, attrib2_name: ....}}
    source_attrib_dict = (
        data[all_cols]
        .assign(node_role=node_role, node_type=column)
        .groupby(column)
        .agg(agg_dict)
        .to_dict(orient="index")
    )
    # we can use the dict to directly set the attributes for all graph items.
    nx.set_node_attributes(graph, source_attrib_dict)


def _pd_unique_list(series: pd.Series):
    """Return either a value or a string if item not unique."""
    unique_vals = series.unique()
    if len(unique_vals) == 1:
        return unique_vals[0]
    return ", ".join([str(attrib) for attrib in unique_vals])


def _verify_columns(
    data, source_col, target_col, source_attrs, target_attrs, edge_attrs
):
    """Check specified columns are in data."""
    missing_columns = {
        **_verify_column(data, "source_col", source_col),
        **_verify_column(data, "target_col", target_col),
    }

    for col in source_attrs or []:
        missing_columns.update(_verify_column(data, "source_attrs", col))
    for col in target_attrs or []:
        missing_columns.update(_verify_column(data, "target_attrs", col))
    for col in edge_attrs or []:
        missing_columns.update(_verify_column(data, "edge_attrs", col))
    if missing_columns:
        raise ValueError(
            "The following parameters had columns that are missing from the input data",
            ", ".join(f"{col} ({param})" for col, param in missing_columns.items()),
        )


def _verify_column(data, param, column):
    """Verify individual column."""
    return {column: param} if column not in data.columns else {}



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""
Polling detection module.

This module is used to highlight edges that are highly periodic and likely to be
generated automatically. The periodic edges could be software polling a server for
updates or malware beaconing and checking for instructions.

There is currently only one technique available for filtering polling data which is
the class PeriodogramPollingDetector.
"""
from collections import Counter
from typing import List, Optional, Tuple, Union

import numpy as np
import numpy.typing as npt
import pandas as pd
from scipy import signal, special

from ..common.utility import export

_PD_VERSION = tuple(int(v) for v in pd.__version__.split("."))
if _PD_VERSION >= (2, 2, 1):
    GROUP_APPLY_PARAMS = {"include_groups": False}
else:
    GROUP_APPLY_PARAMS = {}


@export
class PeriodogramPollingDetector:
    """
    Polling detector using the Periodogram to detect strong frequencies.

    Attributes
    ----------
    data: DataFrame
        Dataframe containing the data to be analysed. Must contain a
        column of edges and a column of timestamps

    """

    def __init__(self, data: pd.DataFrame, copy: bool = False) -> None:
        """
        Create periodogram polling detector.

        Parameters
        ----------
        data: DataFrame
            Dataframe containing the data to be analysed. Must contain a
            column of edges and a column of timestamps

        copy: bool
            A bool to indicate whether to copy the dataframe supplied to data

        """
        if copy:
            self.data = data.copy()
        else:
            self.data = data

    def _g_test(self, pxx: npt.NDArray, exclude_pi: bool) -> Tuple[float, float]:
        """
        Carry out fishers g test for periodicity.

        Fisher's g test tests the null hypothesis that the time series is gaussian white noise
        against the alternative that there is a deterministic periodic component[1]

        If the length of the time series is even then the intensity at pi should be excluded

        If the length of the power spectral density estimate is larger than 700 then an approximate
        p value is calculated otherwise the exact p value is calculate.

        This implementation was taken from the R package GeneCycle[2]

        Parameters
        ----------
        pxx: ArrayLike
            Estimate of the power spectral density

        exclude_pi: bool
            A bool to indicate whether the frequnecy located at pi should be removed.

        Returns
        -------
        Tuple[float, float]
            G test test statistic
            G test P value

        References
        ----------
        [1] M. Ahdesmaki, H. Lahdesmaki and O. Yli-Harja, "Robust Fisher's Test for Periodicity
        Detection in Noisy Biological Time Series," 2007 IEEE International Workshop on Genomic
        Signal Processing and Statistics, 2007, pp. 1-4, doi: 10.1109/GENSIPS.2007.4365817.
        [2] https://github.com/cran/GeneCycle/blob/master/R/fisher.g.test.R

        """
        if exclude_pi:
            pxx = pxx[:-1]

        pxx_length = len(pxx)
        test_statistic = np.max(pxx) / sum(pxx)
        upper = np.floor(1 / test_statistic).astype("int")

        if pxx_length > 700:
            p_value = 1 - (1 - np.exp(-pxx_length * test_statistic)) ** pxx_length
        else:
            compose = []
            for j in range(1, upper):
                compose.append(
                    (-1) ** (j - 1)
                    * np.exp(
                        np.log(special.binom(pxx_length, j))
                        + (pxx_length - 1) * np.log(1 - j * test_statistic)
                    )
                )

            p_value = sum(compose)

        p_value = min(p_value, 1)

        return test_statistic, p_value

    def _detect_polling_arr(
        self,
        timestamps: npt.NDArray,
        process_start: int,
        process_end: int,
        interval: int = 1,
    ) -> Tuple[float, float, float]:
        """
        Carry out periodogram polling detecton on an array of timestamps.

        Carries out the the procedure outlined in [1] to detect if the arrival times have a strong
        periodic component.
        The procedure estimates the periodogram for the data and passes the results to fishers G
        test.

        For more information run PeriodogramPollingDetector._g_test.__doc__

        This code was adapted from [2].

        Parameters
        ----------
        timestamps: ArrayLike
            An array like object containing connection arrival times as timestamps
        process_start: int
            The timestamp representing the start of the counting process
        process_end: int
            The timestamp representing the end of the counting process
        interval: int
            The interval in seconds between observations

        Returns
        -------
        p_val: float
            The p value from fishers G test

        References
        ----------
          [1] Heard, N. A. and Rubin-Delanchy, P. T. G. and Lawson, D. J. (2014) Filtering
          automated polling traffic in computer network flow data. In proceedings of IEEE
          Joint Intelligence and Security Informatics Conference 2014
          [2] https://github.com/fraspass/human_activity/blob/master/fourier.py

        """
        time_steps = np.arange(process_start, process_end, step=interval)
        counting_process = Counter(timestamps)

        dn_ = np.array([counting_process[t] for t in time_steps])
        dn_star = dn_ - len(timestamps) / len(time_steps)

        freq, pxx = signal.periodogram(dn_star)

        max_pxx_freq = freq[np.argmax(pxx)]

        if len(dn_star) % 2 == 0:
            _, p_val = self._g_test(pxx, True)
        else:
            _, p_val = self._g_test(pxx, False)

        return p_val, max_pxx_freq, 1 / max_pxx_freq

    def detect_polling(
        self, time_column: str, groupby: Optional[Union[List[str], str]] = None
    ) -> None:
        """
        Detect the time interval which is highly periodic.

        Runs PeriodogramPollingDetector._detect_polling_arr on the time_column and populates a
        p_value column, dominant_frequency column and dominant_interval column.

        If groupby column(s) are given then PeriodogramPollingDetector._detect_polling_arr is ran on
        each group.

        Parameters
        ----------
        time_column: str
            The name of the column that contains timestamps
        groupby: str or list[str], optional
            Column(s) to group by

        """
        ts_col = self.data[time_column]

        start = min(ts_col)
        end = max(ts_col)

        if not groupby:
            p_value, freq, interval = self._detect_polling_arr(ts_col, start, end)  # type: ignore

            self.data["p_value"] = p_value
            self.data["dominant_frequency"] = freq
            self.data["dominant_interval"] = interval
        else:
            grouped_results = self.data.groupby(groupby).apply(
                lambda x: self._detect_polling_arr(  # type: ignore
                    x[time_column], min(x[time_column]), max(x[time_column])  # type: ignore
                ),
                **GROUP_APPLY_PARAMS,
            )

            grouped_results_df = pd.DataFrame(
                grouped_results.tolist(),
                columns=["p_value", "dominant_frequency", "dominant_interval"],
                index=grouped_results.index,
            ).reset_index()

            self.data = self.data.merge(grouped_results_df)



"""VirusTotal v3 API."""

from __future__ import annotations

import asyncio
import logging
from enum import Enum
from typing import TYPE_CHECKING, Any, ClassVar, Coroutine, Iterable

import pandas as pd
from IPython.core.display import HTML
from IPython.display import display
from typing_extensions import Self

from ...common.exceptions import MsticpyImportExtraError
from ...common.provider_settings import ProviderSettings, get_provider_settings
from ...common.utility import is_ipython

try:
    import nest_asyncio
    import vt

    # Removing dependency temporarily due to build break
    from vt_graph_api import VTGraph
    from vt_graph_api import errors as vt_graph_errs

    from .vtfile_behavior import VTFileBehavior

    if TYPE_CHECKING:
        from vt.iterator import Iterator
        from vt.object import Object

except ImportError as imp_err:
    err_msg: str = (
        "Cannot use this feature without vt-py, vt-graph-api and "
        "nest_asyncio packages installed."
    )
    raise MsticpyImportExtraError(
        err_msg,
        title="Error importing VirusTotal modules.",
        extra="vt3",
    ) from imp_err

logger: logging.Logger = logging.getLogger(__name__)
# pylint: disable=too-many-lines


class MsticpyVTNoDataError(Exception):
    """No data returned from VT API."""


class MsticpyVTGraphSaveGraphError(Exception):
    """Could not save VT Graph."""


class VTEntityType(Enum):
    """VTEntityType: Enum class for VirusTotal entity types."""

    FILE = "file"
    DOMAIN = "domain"
    IP_ADDRESS = "ip_address"
    URL = "url"


class ColumnNames(Enum):
    """Column name enum for DataFrame output."""

    ID = "id"
    TYPE = "type"
    DETECTIONS = "detections"
    SCANS = "scans"
    SOURCE = "source"
    TARGET = "target"
    RELATIONSHIP_TYPE = "relationship_type"
    SOURCE_TYPE = "source_type"
    TARGET_TYPE = "target_type"


class VTObjectProperties(Enum):
    """Enum for VT Object properties."""

    ATTRIBUTES = "attributes"
    RELATIONSHIPS = "relationship"
    LAST_ANALYSIS_STATS = "last_analysis_stats"
    MALICIOUS = "malicious"


def _ensure_eventloop(*, force_nest_asyncio: bool = False) -> asyncio.AbstractEventLoop:
    """Ensure that we have an event loop available."""
    try:
        event_loop: asyncio.AbstractEventLoop = asyncio.get_event_loop()
    except RuntimeError:
        # Generate an event loop if there isn't any.
        event_loop = asyncio.new_event_loop()
        asyncio.set_event_loop(event_loop)
    if is_ipython() or force_nest_asyncio:
        nest_asyncio.apply()
    return event_loop


def _make_sync(future: Coroutine) -> pd.DataFrame:
    """Wait for an async call, making it sync."""
    event_loop: asyncio.AbstractEventLoop = _ensure_eventloop()
    return event_loop.run_until_complete(future)


VT_API_NOT_FOUND = "NotFoundError"


class VTLookupV3:
    """VTLookupV3: VirusTotal lookup of IoC reports."""

    _SUPPORTED_VT_TYPES: ClassVar[set[VTEntityType]] = {
        VTEntityType.FILE,
        VTEntityType.URL,
        VTEntityType.IP_ADDRESS,
        VTEntityType.DOMAIN,
    }

    _MAPPING_TYPES_ENDPOINT: ClassVar[dict[VTEntityType, str]] = {
        VTEntityType.FILE: "files",
        VTEntityType.URL: "urls",
        VTEntityType.IP_ADDRESS: "ip_addresses",
        VTEntityType.DOMAIN: "domains",
    }

    _BASIC_PROPERTIES_PER_TYPE: ClassVar[dict[VTEntityType, set[str]]] = {
        VTEntityType.FILE: {
            "type_description",
            "size",
            "first_submission_date",
            "last_submission_date",
            "times_submitted",
            "meaningful_name",
        },
        VTEntityType.URL: {
            "first_submission_date",
            "last_submission_date",
            "times_submitted",
        },
        VTEntityType.IP_ADDRESS: {"date", "country", "asn", "as_owner"},
        VTEntityType.DOMAIN: {"id", "creation_date", "last_update_date", "country"},
    }

    _DEFAULT_SEARCH_LIMIT: ClassVar[int] = (
        1000  # prevents vague queries from pulling 1M+ files
    )
    _SEARCH_API_ENDPOINT: ClassVar[str] = "/intelligence/search"

    @property
    def supported_vt_types(self: Self) -> list[str]:
        """
        Return list of VirusTotal supported IoC type names.

        Returns
        -------
        list[str]:
            List of VirusTotal supported IoC type names.

        """
        return [str(i_type) for i_type in self._SUPPORTED_VT_TYPES]

    @classmethod
    def _get_endpoint_name(cls: type[Self], vt_type: str) -> str:
        if VTEntityType(vt_type) not in cls._SUPPORTED_VT_TYPES:
            error_msg: str = f"Property type {vt_type} not supported"
            raise KeyError(error_msg)

        return cls._MAPPING_TYPES_ENDPOINT[VTEntityType(vt_type)]

    @classmethod
    def _parse_vt_object(
        cls: type[Self],
        vt_object: vt.object.Object,
        *,
        all_props: bool = False,
    ) -> pd.DataFrame:
        obj_dict: dict[str, Any] = vt_object.to_dict()
        add_columns_df = pd.DataFrame()
        if VTObjectProperties.ATTRIBUTES.value in obj_dict:
            attributes: dict[str, Any] = obj_dict[VTObjectProperties.ATTRIBUTES.value]
            vt_type = VTEntityType(vt_object.type)
            if vt_type not in cls._SUPPORTED_VT_TYPES:
                error_msg: str = f"Property type {vt_type} not supported"
                raise KeyError(error_msg)
            if not all_props:
                obj: dict[str, Any] = {
                    key: attributes[key]
                    for key in cls._BASIC_PROPERTIES_PER_TYPE[vt_type]
                    if key in attributes
                }
            else:
                obj = attributes
            vt_df: pd.DataFrame = pd.json_normalize(data=[obj])
            last_analysis_stats: dict[str, Any] | None = attributes.get(
                VTObjectProperties.LAST_ANALYSIS_STATS.value,
            )
            if last_analysis_stats:
                add_columns_df = pd.DataFrame(
                    {
                        ColumnNames.DETECTIONS.value: [
                            last_analysis_stats[VTObjectProperties.MALICIOUS.value],
                        ],
                        ColumnNames.SCANS.value: [sum(last_analysis_stats.values())],
                    },
                )
            # Format dates for pandas
            vt_df = timestamps_to_utcdate(vt_df)
        elif obj_dict:
            vt_df = pd.json_normalize([obj_dict])
        else:
            vt_df = cls._item_not_found_df(
                vt_type=vt_object.type,
                observable=vt_object.id,
            )

        # Inject ID and Type columns
        if ColumnNames.ID.value not in vt_df.columns:
            add_columns_df[ColumnNames.ID.value] = [vt_object.id]
        if ColumnNames.TYPE.value not in vt_df.columns:
            add_columns_df[ColumnNames.TYPE.value] = [vt_object.type]
        return pd.concat(
            [vt_df, add_columns_df],
            axis="columns",
        )

    def __init__(
        self: VTLookupV3,
        vt_key: str | None = None,
        *,
        force_nestasyncio: bool = False,
    ) -> None:
        """
        Create a new instance of VTLookupV3 class.

        Parameters
        ----------
        vt_key: str, optional
            VirusTotal API key, if not supplied, this is read from
            user configuration.
        force_nestasyncio: bool, optional
            To force use of nest_asyncio, by default nest_asyncio
            is used in Jupyter notebooks, otherwise this defaults to False

        """
        _ensure_eventloop(force_nest_asyncio=force_nestasyncio)
        self._vt_key: str | None = vt_key or _get_vt_api_key()
        self._vt_client = vt.Client(apikey=self._vt_key)

    async def _lookup_ioc_async(
        self: Self,
        observable: str,
        vt_type: str,
        *,
        all_props: bool = False,
    ) -> pd.DataFrame:
        """
        Look up and single IoC observable.

        Parameters
        ----------
        observable: str
            The observable value
        vt_type: str
            The VT entity type
        all_props : bool, optional
            If True, return all properties, by default False

        Returns
        -------
            Attributes Pandas DataFrame with the properties of the entity

        Raises
        ------
        KeyError
            Unknown vt_type

        """
        if VTEntityType(vt_type) not in self._SUPPORTED_VT_TYPES:
            error_msg: str = (
                f"Property type {vt_type} not supported"
                "Valid types are"
                ", ".join(x.value for x in VTEntityType.__members__.values())
            )
            raise KeyError(error_msg)

        endpoint_name: str = self._get_endpoint_name(vt_type)
        try:
            response: Object = self._vt_client.get_object(
                f"/{endpoint_name}/{observable}",
            )
            return self._parse_vt_object(response, all_props=all_props)
        except vt.APIError as err:
            if err.args and err.args[0] == VT_API_NOT_FOUND:
                return self._item_not_found_df(vt_type, observable)
            error_msg = "An error occurred requesting data from VirusTotal"
            raise MsticpyVTNoDataError(error_msg) from err

    def lookup_ioc(
        self: Self,
        observable: str,
        vt_type: str,
        *,
        all_props: bool = False,
    ) -> pd.DataFrame:
        """
        Look up and single IoC observable.

        Parameters
        ----------
        observable: str
            The observable value
        vt_type: str
            The VT entity type
        all_props : bool, optional
            If True, return all properties, by default False

        Returns
        -------
            Attributes Pandas DataFrame with the properties of the entity

        Raises
        ------
        KeyError
            Unknown vt_type

        """
        try:
            return _make_sync(
                self._lookup_ioc_async(observable, vt_type, all_props=all_props),
            )
        finally:
            self._vt_client.close()

    async def _lookup_iocs_async(
        self: Self,
        observables_df: pd.DataFrame,
        observable_column: str = ColumnNames.TARGET.value,
        observable_type_column: str = ColumnNames.TARGET_TYPE.value,
        *,
        all_props: bool = False,
    ) -> pd.DataFrame:
        """
        Look up and multiple IoC observables.

        Parameters
        ----------
        observables_df: pd.DataFrame
            A Pandas DataFrame, where each row is an observable
        observable_column:
            ID column of each observable
        observable_type_column:
            Type column of each observable
        all_props : bool, optional
            If True, return all properties, by default False

        Returns
        -------
            Future Attributes Pandas DataFrame with the properties of the entities

        Raises
        ------
        KeyError
            Column not found in observables_df

        """
        _observables_df: pd.DataFrame = observables_df.reset_index()

        for column in [observable_column, observable_type_column]:
            if column not in _observables_df.columns:
                error_msg: str = f"Column {column} not found in observables_df"
                raise KeyError(error_msg)

        observables_list: pd.Series = _observables_df[observable_column]
        types_list: pd.Series = _observables_df[observable_type_column]
        dfs_futures: list[Coroutine] = [
            self._lookup_ioc_async(
                observable,
                observable_type,
                all_props=all_props,
            )
            for observable, observable_type in zip(observables_list, types_list)
        ]

        dfs: list[pd.DataFrame] = await asyncio.gather(*dfs_futures)
        return (
            pd.concat(dfs)
            if dfs
            else self._item_not_found_df(
                vt_type=", ".join(types_list),
                observable=", ".join(observables_list),
            )
        )

    def lookup_iocs(
        self: Self,
        observables_df: pd.DataFrame,
        observable_column: str = ColumnNames.TARGET.value,
        observable_type_column: str = ColumnNames.TARGET_TYPE.value,
        *,
        all_props: bool = False,
    ) -> pd.DataFrame:
        """
        Look up and multiple IoC observables.

        Parameters
        ----------
        observables_df: pd.DataFrame
            A Pandas DataFrame, where each row is an observable
        observable_column:
            ID column of each observable
        observable_type_column:
            Type column of each observable
        all_props : bool, optional
            If True, return all properties, by default False

        Returns
        -------
            Attributes Pandas DataFrame with the properties of the entities

        """
        try:
            return _make_sync(
                self._lookup_iocs_async(
                    observables_df,
                    observable_column,
                    observable_type_column,
                    all_props=all_props,
                ),
            )
        finally:
            self._vt_client.close()

    async def _lookup_ioc_relationships_async(  # pylint: disable=too-many-locals #noqa: PLR0913
        self: Self,
        observable: str,
        vt_type: str,
        relationship: str,
        *,
        limit: int | None = None,
        all_props: bool = False,
        full_objects: bool = False,
    ) -> pd.DataFrame:
        """
        Look up and single IoC observable relationships.

        Parameters
        ----------
        observable: str
            The observable value
        vt_type: str
            The VT entity type
        relationship: str
            Desired relationship
        limit: int
            Relations limit
        all_props : bool, optional
            If True, return all properties, by default False
        full_objects : bool, optional
            If True, return the full object rather than just ID links.

        Returns
        -------
            Future Relationship Pandas DataFrame with the relationships of the entity

        Raises
        ------
        KeyError
            Unknown vt_type

        """
        if VTEntityType(vt_type) not in self._SUPPORTED_VT_TYPES:
            error_msg: str = f"Property type {vt_type} not supported"
            raise KeyError(error_msg)

        endpoint_name: str = self._get_endpoint_name(vt_type)
        response: vt.object.Object

        if limit is None:
            try:
                response = self._vt_client.get_object(
                    f"/{endpoint_name}/{observable}?relationship_counters=true",
                )
                relationships: dict[str, Any] = response.relationships
                limit = (
                    relationships[relationship]["meta"]["count"]
                    if relationship in relationships
                    else 0
                )
            except KeyError:
                logger.exception(
                    "Could not obtain relationship limit for %s %s",
                    vt_type,
                    observable,
                )
                return self._item_not_found_df(vt_type=vt_type, observable=observable)

        if limit == 0 or limit is None:
            return self._item_not_found_df(vt_type=vt_type, observable=observable)

        obj_relation_type: str = (
            f"{relationship}" if full_objects else f"relationships/{relationship}"
        )
        try:
            relations: Iterable[vt.object.Object] = self._vt_client.iterator(
                f"/{endpoint_name}/{observable}/{obj_relation_type}",
                batch_size=40,
                limit=limit,
            )
            vt_objects: list[pd.DataFrame] = [
                self._parse_vt_object(r, all_props=all_props) for r in relations
            ]
            result_df: pd.DataFrame = (
                pd.concat(vt_objects, ignore_index=True)
                if vt_objects
                else self._relation_not_found_df(
                    vt_type=vt_type,
                    observable=observable,
                    relationship=relationship,
                )
            )

            if vt_objects:
                # Add additional columns (with same size as DF)
                rows: int = len(result_df)
                add_columns = pd.DataFrame(
                    {
                        ColumnNames.SOURCE.value: [observable] * rows,
                        ColumnNames.SOURCE_TYPE.value: [VTEntityType(vt_type).value]
                        * rows,
                        ColumnNames.RELATIONSHIP_TYPE.value: [relationship] * rows,
                    },
                )

                result_df = (
                    pd.concat([result_df, add_columns], axis="columns")
                    .rename(
                        columns={
                            ColumnNames.ID.value: ColumnNames.TARGET.value,
                            ColumnNames.TYPE.value: ColumnNames.TARGET_TYPE.value,
                        },
                    )
                    .reset_index()
                )
        except vt.APIError as err:
            if err.args and err.args[0] == VT_API_NOT_FOUND:
                return self._relation_not_found_df(vt_type, observable, relationship)
            error_msg = "An error occurred requesting data from VirusTotal"
            raise MsticpyVTNoDataError(error_msg) from err

        return result_df

    def lookup_ioc_relationships(  # noqa: PLR0913
        self: Self,
        observable: str,
        vt_type: str,
        relationship: str,
        *,
        limit: int | None = None,
        all_props: bool = False,
    ) -> pd.DataFrame:
        """
        Look up single IoC observable relationship links.

        Parameters
        ----------
        observable: str
            The observable value
        vt_type: str
            The VT entity type
        relationship: str
            Desired relationship
        limit: int
            Relations limit
        all_props : bool, optional
            If True, return all properties, by default False

        Returns
        -------
            Relationship Pandas DataFrame with the relationships of the entity

        Notes
        -----
        This method returns relationship links rather than whole objects.
        That is, it will return the IDs of related items in the specified
        `relationship`, if any.

        See Also
        --------
        lookup_ioc_related : return the full related objects.

        """
        try:
            return _make_sync(
                self._lookup_ioc_relationships_async(
                    observable,
                    vt_type,
                    relationship,
                    limit=limit,
                    all_props=all_props,
                ),
            )
        finally:
            self._vt_client.close()

    def lookup_ioc_related(
        self: Self,
        observable: str,
        vt_type: str,
        relationship: str,
        *,
        limit: int | None = None,
    ) -> pd.DataFrame:
        """
        Look single IoC observable related items.

        Parameters
        ----------
        observable: str
            The observable value
        vt_type: str
            The VT entity type
        relationship: str
            Desired relationship
        limit: int
            Relations limit

        Returns
        -------
        pd.DataFrame
            Any objects with specified `relationship` to the entity

        Notes
        -----
        This method returns full related objects rather than ID links.
        It is less efficient than looking up ID links only.

        See Also
        --------
        lookup_ioc_relationships : return the related IDs.

        """
        try:
            return _make_sync(
                self._lookup_ioc_relationships_async(
                    observable,
                    vt_type,
                    relationship,
                    limit=limit,
                    all_props=True,
                    full_objects=True,
                ),
            )
        finally:
            self._vt_client.close()

    async def _lookup_iocs_relationships_async(  # noqa: PLR0913
        self: Self,
        observables_df: pd.DataFrame,
        relationship: str,
        observable_column: str = ColumnNames.TARGET.value,
        observable_type_column: str = ColumnNames.TARGET_TYPE.value,
        *,
        limit: int | None = None,
        all_props: bool = False,
    ) -> pd.DataFrame:
        """
        Look up and single IoC observable relationships.

        Parameters
        ----------
        observables_df: pd.DataFrame
            A Pandas DataFrame, where each row is an observable
        relationship: str
            Desired relationship
        observable_column:
            ID column of each observable
        observable_type_column:
            Type column of each observable.
        limit: int
            Relations limit
        all_props : bool, optional
            If True, return all properties, by default False

        Returns
        -------
            Future Relationship Pandas DataFrame with the relationships
            of each observable.

        Raises
        ------
        KeyError
            Column not found in observables_df

        """
        _observables_df: pd.DataFrame = observables_df.reset_index()

        for column in [observable_column, observable_type_column]:
            if column not in _observables_df.columns:
                error_msg: str = f"Column {column} not found in observables df"
                raise KeyError(error_msg)

        observables_list: pd.Series = _observables_df[observable_column]
        types_list: pd.Series = _observables_df[observable_type_column]
        dfs_futures: list[Coroutine[Any, Any, pd.DataFrame]] = [
            self._lookup_ioc_relationships_async(
                observable,
                observable_type,
                relationship,
                limit=limit,
                all_props=all_props,
            )
            for observable, observable_type in zip(observables_list, types_list)
        ]

        dfs: list[pd.DataFrame] = await asyncio.gather(*dfs_futures)
        return (
            pd.concat(dfs)
            if len(dfs) > 0
            else self._relation_not_found_df(
                vt_type=", ".join(types_list),
                observable=", ".join(observables_list),
                relationship=relationship,
            )
        )

    def lookup_iocs_relationships(  # noqa: PLR0913
        self: Self,
        observables_df: pd.DataFrame,
        relationship: str,
        observable_column: str = ColumnNames.TARGET.value,
        observable_type_column: str = ColumnNames.TARGET_TYPE.value,
        *,
        limit: int | None = None,
        all_props: bool = False,
    ) -> pd.DataFrame:
        """
        Look up and single IoC observable relationships.

        Parameters
        ----------
        observables_df: pd.DataFrame
            A Pandas DataFrame, where each row is an observable
        relationship: str
            Desired relationship
        observable_column:
            ID column of each observable
        observable_type_column:
            Type column of each observable.
        limit: int
            Relations limit
        all_props : bool, optional
            If True, return all properties, by default False

        Returns
        -------
            Relationship Pandas DataFrame with the relationships of each observable.

        """
        try:
            return _make_sync(
                self._lookup_iocs_relationships_async(
                    observables_df,
                    relationship,
                    observable_column,
                    observable_type_column,
                    limit=limit,
                    all_props=all_props,
                ),
            )

        finally:
            self._vt_client.close()

    def create_vt_graph(
        self: Self,
        relationship_dfs: list[pd.DataFrame],
        name: str,
        *,
        private: bool,
    ) -> str:
        """
        Create a VirusTotal Graph with a set of Relationship DataFrames.

        Parameters
        ----------
        relationship_dfs:
            List of Relationship DataFrames
        name:
            New graph name
        private
            Indicates if the Graph is private or not.

        Returns
        -------
            Graph ID

        Raises
        ------
            ValueError when private is not indicated.
            ValueError when there are no relationship DataFrames
            MsticpyVTGraphSaveGraphError when Graph can not be saved

        """
        if not isinstance(private, bool):
            error_msg = "Please indicate if Graph is private or not"
            raise TypeError(error_msg)

        nodes, edges = self.relationships_to_graph(relationship_dfs)
        graph = VTGraph(self._vt_key, name=name, private=private)
        graph.add_nodes(nodes)

        for edge in edges:
            graph.add_link(**edge)

        try:
            graph.save_graph()
        except vt_graph_errs.SaveGraphError as graph_err:
            graph_mssg: list[str] = (
                [
                    "Please check you have Private Graph premium feature enabled in",
                    "your subscription. It is possible to create public Graphs",
                    "with 'private=False' input argument",
                ]
                if private
                else [""]
            )
            error_msg = "Could not save Graph."
            raise MsticpyVTGraphSaveGraphError(
                error_msg,
                *graph_mssg,
            ) from graph_err

        return graph.graph_id

    def get_object(self: Self, vt_id: str, vt_type: str) -> pd.DataFrame:
        """
        Return the full VT object as a DataFrame.

        Parameters
        ----------
        vt_id : str
            The ID of the object
        vt_type : str
            The type of object to query.

        Returns
        -------
        pd.DataFrame
            Single column DataFrame with attribute names as
            index and values as data column.

        Raises
        ------
        KeyError
            Unrecognized VT Type
        MsticpyVTNoDataError
            Error requesting data from VT.

        Notes
        -----
        This calls the underlying VT get_object API directly and
        returns all attributes for the object - hence a very wide
        DataFrame.

        """
        if VTEntityType(vt_type) not in self._SUPPORTED_VT_TYPES:
            # pylint: disable=no-member

            error_msg: str = (
                f"Property type {vt_type} not supported. "
                "Valid types are: "
                ", ".join(x.value for x in VTEntityType.__members__.values())
            )
            raise KeyError(error_msg)

        endpoint_name = self._get_endpoint_name(vt_type)
        try:
            response: vt.object.Object = self._vt_client.get_object(
                f"/{endpoint_name}/{vt_id}",
            )
            result_df = pd.DataFrame(
                {
                    "id": [response.id],
                    "type": [response.type],
                },
            )
            attribs = pd.json_normalize(response.to_dict()["attributes"])
            result_df = pd.concat([result_df, attribs], axis=1)
            result_df["context_attributes"] = response.to_dict().get(
                "context_attributes",
            )
            return timestamps_to_utcdate(result_df)
        except vt.APIError as err:
            if err.args and err.args[0] == VT_API_NOT_FOUND:
                return self._item_not_found_df(vt_type, vt_id)
            error_msg = "An error occurred requesting data from VirusTotal"
            raise MsticpyVTNoDataError(error_msg) from err
        finally:
            self._vt_client.close()

    def get_file_behavior(
        self: Self,
        file_id: str | None = None,
        file_summary: dict[str, Any] | None = None,
        sandbox: str | None = None,
    ) -> VTFileBehavior:
        """
        Return a VTFileBehavior object with file detonation results.

        Parameters
        ----------
        file_id : Optional[str], optional
            The ID of the file to look up, by default None
        file_summary : Optional[dict[str, Any]], optional
            VT file summary object dictionary, by default None
        sandbox : str, optional
            Name of specific sandbox to retrieve, by default None
            If None, it will retrieve the behavior summary.

        Returns
        -------
        VTFileBehavior

        """
        if not self._vt_key:
            error_msg: str = "VT key is required to retrieve file behavior"
            raise ValueError(error_msg)
        vt_behavior = VTFileBehavior(
            self._vt_key,
            file_id=file_id,
            file_summary=file_summary,
        )
        vt_behavior.get_file_behavior(sandbox=sandbox)
        return vt_behavior

    def search(
        self: Self,
        query: str,
        limit: int = _DEFAULT_SEARCH_LIMIT,
    ) -> pd.DataFrame:
        """
        Return results of a VT search query as a DataFrame.

        Parameters
        ----------
        query : str
            Virus Total Intelligence Search string
        limit : int (default: 1,000)
            Number of intended results

        Returns
        -------
        pd.DataFrame
            Search query results.

        """
        # run virus total intelligence search using iterator
        try:
            response_itr: Iterator = self._vt_client.iterator(
                self._SEARCH_API_ENDPOINT,
                params={"query": query},
                limit=limit,
            )
            response_list: list[dict[str, Any]] = [
                item.to_dict() for item in response_itr
            ]
        except vt.APIError as api_err:
            error_msg: str = (
                f"The provided query returned 0 results because of an APIError: {api_err}"
            )
            raise MsticpyVTNoDataError(error_msg) from api_err

        if len(response_list) == 0:
            error_msg = "The provided query returned 0 results"
            raise MsticpyVTNoDataError(error_msg)

        response_df: pd.DataFrame = self._extract_response(response_list)
        return timestamps_to_utcdate(response_df)

    def iterator(  # noqa: PLR0913
        self: Self,
        path: str,
        *path_args: str,
        params: dict[str, Any] | None = None,
        cursor: str | None = None,
        limit: int | None = None,
        batch_size: int = 0,
    ) -> vt.Iterator:
        """
        Return an iterator for the collection specified by the given path.

        The endpoint specified by path must return a collection of objects. An
        example of such an endpoint are /comments and /intelligence/search.

        SOURCE: https://github.com/VirusTotal/vt-py/blob/
        6bf4decb5bbd80bfc60e74ee3caa4c9073cea38c/vt/client.py

        Parameters
        ----------
        path : str
            Path to API endpoint returning a collection.
        path_args: dict
            A variable number of arguments that are put
            into any placeholders used in path.
        params: dict
            Additional parameters passed to the endpoint.
        cursor: str
            Cursor for resuming the iteration at the point it was left previously.
            A cursor can be obtained with Iterator.cursor(). This
            cursor is not the same one returned by the VirusTotal API.
        limit: int
            Maximum number of objects that will be returned by the iterator.
            If a limit is not provided the iterator continues until it reaches the
            last object in the collection.
        batch_size: int
            Maximum number of objects retrieved on each call to the endpoint.
            If not provided the server will decide how many objects to return.

        Returns
        -------
        vt.Iterator
            An instance of vt.Iterator

        """
        return self._vt_client.iterator(
            path,
            path_args,
            params,
            cursor,
            limit,
            batch_size,
        )

    def _extract_response(self: Self, response_list: list) -> pd.DataFrame:
        """
        Convert list of dictionaries from search() function to DataFrame.

        Parameters
        ----------
        response_list : list
            A list of dictionaries representing a virus total object

        Returns
        -------
        pd.DataFrame
            A DataFrame with the attributes of the virus total object

        """
        # loop to convert from list of vt.Object to pd.DataFrame
        response_rows = []
        for response_item in response_list:
            # flatten nested dictionary and append id, type values
            response_item_df = pd.json_normalize(response_item["attributes"])
            response_item_df["id"] = response_item["id"]
            response_item_df["type"] = response_item["type"]

            response_rows.append(response_item_df)

        response_df = pd.concat(response_rows, axis=0, ignore_index=True)
        return timestamps_to_utcdate(response_df)

    @staticmethod
    def relationships_to_graph(
        relationship_dfs: list[pd.DataFrame],
    ) -> tuple[list[dict[str, Any]], list[dict[str, Any]]]:
        """
        Generate nodes and edges from relationships.

        Parameters
        ----------
        relationship_dfs : list[pd.DataFrame]
            List of relationship DataFrames

        Returns
        -------
        Tuple[list[dict[str, Any]], list[dict[str, Any]]]
            List of nodes (node_id, node_type)
            List of edges (source_node, target_node, connection_type)

        Raises
        ------
        ValueError
            If an empty list is supplied.

        """
        if not relationship_dfs:
            error_msg: str = "There are no relationship DataFrames"
            raise ValueError(error_msg)

        # Set indexes to source-target so that pd.concat aligns on these indexes
        relationship_dfs = [
            rel_df.set_index([ColumnNames.SOURCE.value, ColumnNames.TARGET.value])
            for rel_df in relationship_dfs
        ]
        concatenated_df = pd.concat(relationship_dfs).reset_index()

        # Create nodes DF, with source and target
        sources_df = (
            concatenated_df.groupby(ColumnNames.SOURCE.value)[
                ColumnNames.SOURCE_TYPE.value
            ]
            .first()
            .reset_index()
            .rename(
                columns={
                    ColumnNames.SOURCE.value: ColumnNames.ID.value,
                    ColumnNames.SOURCE_TYPE.value: ColumnNames.TYPE.value,
                },
            )
        )

        target_df = (
            concatenated_df.groupby(ColumnNames.TARGET.value)[
                ColumnNames.TARGET_TYPE.value
            ]
            .first()
            .reset_index()
            .rename(
                columns={
                    ColumnNames.TARGET.value: ColumnNames.ID.value,
                    ColumnNames.TARGET_TYPE.value: ColumnNames.TYPE.value,
                },
            )
        )

        nodes_df = pd.concat([sources_df, target_df])
        nodes = [
            {
                "node_id": row[ColumnNames.ID.value],
                "node_type": row[ColumnNames.TYPE.value],
            }
            for row in nodes_df.itertuples(index=False)
        ]

        edges = [
            {
                "source_node": row[ColumnNames.SOURCE.value],
                "target_node": row[ColumnNames.TARGET.value],
                "connection_type": row[ColumnNames.RELATIONSHIP_TYPE.value],
            }
            for row in concatenated_df.itertuples(index=False)
        ]
        return nodes, edges

    @staticmethod
    def render_vt_graph(
        graph_id: str,
        width: int = 800,
        height: int = 600,
    ) -> None:
        """
        Display a VTGraph in a Jupyter Notebook.

        Parameters
        ----------
        graph_id:
            Graph ID
        width
            Graph width.
        height
            Graph height

        """
        display(
            HTML(
                f"""
              <iframe
                src="https://www.virustotal.com/graph/embed/{graph_id}"
                width="{width}"
                height="{height}">
              </iframe>

            """,
            ),
        )

    @classmethod
    def _item_not_found_df(
        cls: type[Self],
        vt_type: str,
        observable: str,
    ) -> pd.DataFrame:
        not_found_dict: dict[str, str] = {
            ColumnNames.ID.value: observable,
            ColumnNames.TYPE.value: vt_type,
        }
        vte_type = VTEntityType(vt_type)
        if vte_type not in cls._SUPPORTED_VT_TYPES:
            not_found_dict["status"] = "Unsupported type"
        else:
            not_found_dict.update(
                {key: "Not found" for key in cls._BASIC_PROPERTIES_PER_TYPE[vte_type]},
            )
        return pd.DataFrame([not_found_dict])

    @classmethod
    def _relation_not_found_df(
        cls: type[Self],
        vt_type: str,
        observable: str,
        relationship: str,
    ) -> pd.DataFrame:
        not_found_dict: dict[str, str] = {
            ColumnNames.SOURCE.value: observable,
            ColumnNames.SOURCE_TYPE.value: vt_type,
            ColumnNames.RELATIONSHIP_TYPE.value: relationship,
            ColumnNames.TARGET.value: "Not found",
            ColumnNames.TARGET_TYPE.value: "Not found",
        }
        return pd.DataFrame([not_found_dict])


def _get_vt_api_key() -> str | None:
    """Retrieve the VT key from settings."""
    prov_settings: dict[str, ProviderSettings] = get_provider_settings("TIProviders")
    vt_settings: ProviderSettings | None = prov_settings.get("VirusTotal")
    if vt_settings:
        return vt_settings.args.get("AuthKey")
    return None


def timestamps_to_utcdate(data: pd.DataFrame) -> pd.DataFrame:
    """Replace Unix timestamps in VT data with Py/pandas Timestamp."""
    columns: pd.Index = data.columns
    for date_col in (
        col for col in columns if isinstance(col, str) and col.endswith("_date")
    ):
        data = (
            data.assign(pd_data=pd.to_datetime(data[date_col], unit="s", utc=True))
            .drop(columns=date_col)
            .rename(columns={"pd_data": date_col})
        )
    return data[columns]



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""KQL Driver class."""

import contextlib
import json
import logging
import os
import re
import warnings
from typing import Any, Dict, List, Optional, Tuple, Union

import pandas as pd
from azure.core.exceptions import ClientAuthenticationError
from IPython.core.getipython import get_ipython

from ..._version import VERSION
from ...auth.azure_auth import AzureCloudConfig, az_connect
from ...auth.azure_auth_core import only_interactive_cred
from ...common.exceptions import (
    MsticpyDataQueryError,
    MsticpyImportExtraError,
    MsticpyKqlConnectionError,
    MsticpyNoDataSourceError,
    MsticpyNotConnectedError,
)
from ...common.utility import MSTICPY_USER_AGENT, export
from ...common.wsconfig import WorkspaceConfig
from ..core.query_defns import DataEnvironment
from .driver_base import DriverBase, DriverProps, QuerySource

_KQL_ENV_OPTS = "KQLMAGIC_CONFIGURATION"


# Need to set KQL option before importing
def _set_kql_env_option(option, value):
    """Set an item in the KqlMagic main config environment variable."""
    kql_config = os.environ.get(_KQL_ENV_OPTS, "")
    current_opts = {
        opt.split("=")[0].strip(): opt.split("=")[1]
        for opt in kql_config.split(";")
        if opt.strip() and "=" in opt
    }

    current_opts[option] = value
    kql_config = ";".join(f"{opt}={val}" for opt, val in current_opts.items())
    os.environ[_KQL_ENV_OPTS] = kql_config


_set_kql_env_option("enable_add_items_to_help", False)

try:
    from Kqlmagic import kql as kql_exec
    from Kqlmagic.kql_engine import KqlEngineError
    from Kqlmagic.kql_proxy import KqlResponse
    from Kqlmagic.kql_response import KqlError
    from Kqlmagic.my_aad_helper import AuthenticationError
except ImportError as imp_err:
    raise MsticpyImportExtraError(
        "Cannot use this feature without Kqlmagic installed",
        "Install msticpy with the [kql] extra or one of the following:",
        "%pip install Kqlmagic   # notebook",
        "python -m pip install Kqlmagic   # python",
        title="Error importing Kqlmagic",
        extra="kql",
    ) from imp_err

__version__ = VERSION
__author__ = "Ian Hellen"

_KQL_CLOUD_MAP = {"global": "public", "cn": "china", "usgov": "government"}

_KQL_OPTIONS = ["timeout"]
_KQL_ENV_OPTS = "KQLMAGIC_CONFIGURATION"

_AZ_CLOUD_MAP = {kql_cloud: az_cloud for az_cloud, kql_cloud in _KQL_CLOUD_MAP.items()}

# pylint: disable=too-many-instance-attributes


@export
class KqlDriver(DriverBase):
    """KqlDriver class to execute kql queries."""

    def __init__(self, connection_str: str = None, **kwargs):
        """
        Instantiate KqlDriver and optionally connect.

        Parameters
        ----------
        connection_str : str, optional
            Connection string

        Other Parameters
        ----------------
        debug : bool
            print out additional diagnostic information.

        """
        self.az_cloud_config = AzureCloudConfig()
        self._ip = get_ipython()
        self._debug = kwargs.get("debug", False)
        super().__init__(**kwargs)
        self.workspace_id: Optional[str] = None
        self._loaded = self._is_kqlmagic_loaded()

        os.environ["KQLMAGIC_LOAD_MODE"] = "silent"
        if not self._loaded:
            self._load_kql_magic()

        self._set_kql_option("request_user_agent_tag", MSTICPY_USER_AGENT)
        self._set_kql_env_option("enable_add_items_to_help", False)
        self._schema: Dict[str, Any] = {}
        self.environment = kwargs.pop("data_environment", DataEnvironment.MSSentinel)
        self.set_driver_property(
            DriverProps.EFFECTIVE_ENV, DataEnvironment.MSSentinel.name
        )
        self.kql_cloud, self.az_cloud = self._set_kql_cloud()
        for option, value in kwargs.items():
            self._set_kql_option(option, value)

        self.current_connection = ""
        self.current_connection_args: Dict[str, Any] = {}
        if connection_str:
            self.current_connection = connection_str
            self.current_connection_args.update(kwargs)
            self.connect(connection_str)

    # pylint: disable=too-many-branches
    def connect(self, connection_str: Optional[str] = None, **kwargs):  # noqa: MC0001
        """
        Connect to data source.

        Parameters
        ----------
        connection_str : Union[str, WorkspaceConfig, None]
            Connection string or WorkspaceConfig for the Sentinel Workspace.

        Other Parameters
        ----------------
        kqlmagic_args : str, optional
            Additional string of parameters to be passed to KqlMagic
        mp_az_auth : Union[bool, str, list, None], optional
            Optional parameter directing KqlMagic to use MSTICPy Azure authentication.
            Values can be:
            True or "default": use the settings in msticpyconfig.yaml 'Azure' section
            str: single auth method name
            ('msi', 'cli', 'env', 'vscode', 'powershell', 'cache' or 'interactive')
            List[str]: list of acceptable auth methods from
            ('msi', 'cli', 'env', 'vscode', 'powershell', 'cache' or 'interactive')
        mp_az_tenant_id: str, optional
            Optional parameter specifying a Tenant ID for use by MSTICPy Azure
            authentication.
        workspace : str, optional
            Alternative to supplying a WorkspaceConfig object as the connection_str
            parameter. Giving a workspace name will fetch the workspace
            settings from msticpyconfig.yaml.


        """
        if not self._previous_connection:
            print("Connecting...", end=" ")

        mp_az_auth = kwargs.get("mp_az_auth", "default")
        mp_az_tenant_id = kwargs.get("mp_az_tenant_id")
        workspace = kwargs.get("workspace")
        if workspace or connection_str is None:
            connection_str = WorkspaceConfig(workspace=workspace)  # type: ignore

        if isinstance(connection_str, WorkspaceConfig):
            if not mp_az_tenant_id and "tenant_id" in connection_str:
                mp_az_tenant_id = connection_str["tenant_id"]
            self._instance = connection_str.workspace_key
            connection_str = connection_str.code_connect_str

        if not connection_str:
            raise MsticpyKqlConnectionError(
                f"A connection string is needed to connect to {self._connect_target}",
                title="no connection string",
            )
        if "kqlmagic_args" in kwargs:
            connection_str = f"{connection_str} {kwargs['kqlmagic_args']}"

        # Default to using Azure Auth if possible.
        if mp_az_auth and "try_token" not in kwargs:
            self._set_az_auth_option(mp_az_auth, mp_az_tenant_id)

        self.current_connection = connection_str
        ws_in_connection = re.search(
            r"workspace\(['\"]([^'\"]+).*",
            self.current_connection,
            re.IGNORECASE,
        )
        self.workspace_id = ws_in_connection[1] if ws_in_connection else None
        self.current_connection_args.update(kwargs)
        kql_err_setting = self._get_kql_option("short_errors")
        self._connected = False
        try:
            self._set_kql_option("short_errors", False)
            if self._ip is not None:
                try:
                    kql_exec(connection_str)
                    if not self._previous_connection:
                        print("connected")
                except KqlError as ex:
                    self._raise_kql_error(ex)
                except KqlEngineError as ex:
                    self._raise_kql_engine_error(ex)
                except AuthenticationError as ex:
                    self._raise_authn_error(ex)
                except Exception as ex:  # pylint: disable=broad-except
                    self._raise_adal_error(ex)
                self._connected = True
                self._previous_connection = True
                self._schema = self._get_schema()
            else:
                print(f"Could not connect to kql query provider for {connection_str}")
            return self._connected
        finally:
            self._set_kql_option("short_errors", kql_err_setting)

    # pylint: disable=too-many-branches

    @property
    def schema(self) -> Dict[str, Dict]:
        """
        Return current data schema of connection.

        Returns
        -------
        Dict[str, Dict]
            Data schema of current connection.

        """
        return self._schema

    def query(
        self, query: str, query_source: QuerySource = None, **kwargs
    ) -> Union[pd.DataFrame, Any]:
        """
        Execute query string and return DataFrame of results.

        Parameters
        ----------
        query : str
            The query to execute
        query_source : QuerySource
            The query definition object

        Returns
        -------
        Union[pd.DataFrame, results.ResultSet]
            A DataFrame (if successful) or
            the underlying provider result if an error.

        """
        if query_source:
            try:
                table = query_source["args.table"]
            except KeyError:
                table = None
            if table:
                if " " in table.strip():
                    table = table.strip().split(" ")[0]
                if table not in self.schema:
                    raise MsticpyNoDataSourceError(
                        f"The table {table} for this query is not in your workspace",
                        " or database schema. Please check your this",
                        title=f"{table} not found.",
                    )
        data, result = self.query_with_results(query, **kwargs)
        return data if data is not None else result

    # pylint: disable=too-many-branches
    def query_with_results(
        self, query: str, **kwargs
    ) -> Tuple[pd.DataFrame, KqlResponse]:
        """
        Execute query string and return DataFrame of results.

        Parameters
        ----------
        query : str
            The kql query to execute

        Returns
        -------
        Tuple[pd.DataFrame, results.ResultSet]
            A DataFrame (if successful) and
            Kql ResultSet.

        """
        debug = kwargs.pop("debug", self._debug)
        if debug:
            print(query)

        if (
            not self.connected
            or self.workspace_id != self._get_kql_current_connection()
        ):
            self._make_current_connection()

        # save current auto_dataframe setting so that we can set to false
        # and restore current setting
        auto_dataframe = self._get_kql_option(option="auto_dataframe")
        self._set_kql_option(option="auto_dataframe", value=False)

        # run the query (append semicolon to prevent default output)
        if not query.strip().endswith(";"):
            query = f"{query}\n;"

        # Add any Kqlmagic options from kwargs
        kql_opts = {
            option: option_val
            for option, option_val in kwargs.items()
            if option in _KQL_OPTIONS
        }
        result = kql_exec(query, options=kql_opts)
        self._set_kql_option(option="auto_dataframe", value=auto_dataframe)
        if result is not None:
            if isinstance(result, pd.DataFrame):
                return result, None
            if hasattr(result, "completion_query_info") and (
                int(result.completion_query_info.get("StatusCode", 1)) == 0
                or result.completion_query_info.get("Text")
                == "Query completed successfully"
            ):
                data_frame = result.to_dataframe()
                if result.is_partial_table:
                    print("Warning - query returned partial results.")
                if debug:
                    print("Query status:\n", "\n".join(self._get_query_status(result)))
                return data_frame, result

        return self._raise_query_failure(query, result)

    def _make_current_connection(self):
        """Switch to the current connection (self.current_connection)."""
        try:
            self.connect(self.current_connection, **(self.current_connection_args))
        except MsticpyKqlConnectionError:
            self._connected = False
        if not self.connected:
            raise MsticpyNotConnectedError(
                "Please run the connect() method before running a query.",
                title=f"not connected to a {self._connect_target}",
                help_uri=MsticpyKqlConnectionError.DEF_HELP_URI,
            )

    def _load_kql_magic(self):
        """Load KqlMagic if not loaded."""
        # KqlMagic
        print("Please wait. Loading Kqlmagic extension...", end="")
        if self._ip is not None:
            with warnings.catch_warnings():
                # Suppress logging exception about PyGObject from msal_extensions
                msal_ext_logger = logging.getLogger("msal_extensions.libsecret")
                current_level = msal_ext_logger.getEffectiveLevel()
                msal_ext_logger.setLevel(logging.CRITICAL)
                warnings.simplefilter(action="ignore")
                self._ip.run_line_magic("reload_ext", "Kqlmagic")
                msal_ext_logger.setLevel(current_level)
        self._loaded = True
        print("done")

    def _is_kqlmagic_loaded(self) -> bool:
        """Return true if kql magic is loaded."""
        if self._ip is not None:
            return self._ip.find_magic("kql") is not None
        return bool(kql_exec("--version"))

    @property
    def _connect_target(self) -> str:
        if self.environment == DataEnvironment.MSSentinel:
            return "Workspace"
        return "Kusto cluster"

    @staticmethod
    def _get_query_status(result) -> List[str]:
        return [f"{key}: '{value}'" for key, value in result.completion_query_info]

    @staticmethod
    def _get_schema() -> Dict[str, Dict]:
        return kql_exec("--schema")

    @staticmethod
    def _get_kql_option(option):
        """Retrieve a current Kqlmagic notebook option."""
        return kql_exec(f"--config {option}").get(option)

    @staticmethod
    def _set_kql_option(option, value):
        """Set a Kqlmagic notebook option."""
        kql_exec("--config short_errors=False")
        result: Any
        try:
            opt_val = f"'{value}'" if isinstance(value, str) else value
            result = kql_exec(f"--config {option}={opt_val}")
        except ValueError:
            result = None
        finally:
            kql_exec("--config short_errors=True")
        return result

    @staticmethod
    def _set_kql_env_option(option, value):
        """Set an item in the KqlMagic main config environment variable."""
        kql_config = os.environ.get(_KQL_ENV_OPTS, "")
        current_opts = {
            opt.split("=")[0].strip(): opt.split("=")[1]
            for opt in kql_config.split(";")
        }
        current_opts[option] = value
        kql_config = ";".join(f"{opt}={val}" for opt, val in current_opts.items())
        # print(kql_config) - replace with logger
        os.environ[_KQL_ENV_OPTS] = kql_config

    @staticmethod
    def _get_kql_current_connection():
        """Get the current connection Workspace ID from KQLMagic."""
        connections = kql_exec("--conn")
        current_connection = [conn for conn in connections if conn.startswith(" * ")]
        if not current_connection:
            return ""
        return current_connection[0].strip(" * ").split("@")[0]

    def _set_kql_cloud(self):
        """If cloud is set in Azure Settings override default."""
        # Check that there isn't a cloud setting in the KQLMAGIC env var
        kql_config = os.environ.get(_KQL_ENV_OPTS, "")
        if "cloud" in kql_config:
            # Set by user - we don't want to override this
            kql_cloud = self._get_kql_option("cloud")
            az_cloud = _AZ_CLOUD_MAP.get(kql_cloud, "public")
            return kql_cloud, az_cloud
        az_cloud = self.az_cloud_config.cloud
        kql_cloud = _KQL_CLOUD_MAP.get(az_cloud, "public")
        if kql_cloud != self._get_kql_option("cloud"):
            self._set_kql_option("cloud", kql_cloud)
        return kql_cloud, az_cloud

    @staticmethod
    def _raise_query_failure(query, result):
        """Raise query failure exception."""
        err_contents = []
        if hasattr(result, "completion_query_info"):
            q_info = result.completion_query_info
            if "StatusDescription" in q_info:
                err_contents = [
                    f"StatusDescription {q_info.get('StatusDescription')}",
                    f"(err_code: {result.completion_query_info.get('StatusCode')})",
                ]
            elif "Text" in q_info:
                err_contents = [f"StatusDescription {q_info.get('Text')}"]
            else:
                err_contents = [f"Unknown error type: {q_info}"]
        if not err_contents:
            err_contents = ["Unknown query error"]

        err_contents.append(f"Query:\n{query}")
        raise MsticpyDataQueryError(*err_contents)

    _WS_RGX = r"workspace\(['\"](?P<ws>[^'\"]+)"
    _TEN_RGX = r"tenant\(['\"](?P<tenant>[^'\"]+)"

    def _raise_kql_error(self, ex):
        kql_err = json.loads(ex.args[0]).get("error")
        if kql_err.get("code") == "WorkspaceNotFoundError":
            ex_mssgs = [
                "The workspace ID used to connect to Microsoft Sentinel could not be found.",
                "Please check that this is a valid workspace for your subscription",
            ]
            ws_match = re.search(self._WS_RGX, self.current_connection, re.IGNORECASE)
            if ws_match:
                ws_name = ws_match.groupdict().get("ws")
                ex_mssgs.append(f"The workspace id used was {ws_name}.")
            ex_mssgs.append(f"The full connection string was {self.current_connection}")
            raise MsticpyKqlConnectionError(*ex_mssgs, title="unknown workspace")
        raise MsticpyKqlConnectionError(
            "The service returned the following error when connecting",
            str(ex),
            title="Kql response error",
        )

    @staticmethod
    def _raise_kql_engine_error(ex):
        ex_mssgs = [
            "An error was returned from Kqlmagic KqlEngine.",
            "This can occur if you tried to connect to a second workspace using a"
            + " different tenant ID - only a single tenant ID is supported in"
            + " one notebook.",
            "Other causes of this error could be an invalid format of your"
            + " connection string",
            *(ex.args),
        ]
        raise MsticpyKqlConnectionError(*ex_mssgs, title="kql connection error")

    @staticmethod
    def _raise_adal_error(ex):
        """Adal error - usually wrong tenant ID."""
        if ex.args and ex.args[0] == "Unexpected polling state code_expired":
            raise MsticpyKqlConnectionError(
                "Authentication request was not completed.",
                title="authentication timed out",
            )

        err_response = getattr(ex, "error_response", None)
        if err_response and "error_description" in ex.error_response:
            ex_mssgs = ex.error_response["error_description"].split("\r\n")
        else:
            ex_mssgs = [f"Full error: {ex}"]
        raise MsticpyKqlConnectionError(
            *ex_mssgs, title="could not authenticate to tenant"
        )

    @staticmethod
    def _raise_authn_error(ex):
        """Raise an authentication error."""
        ex_mssgs = [
            "The authentication failed.",
            "Please check the credentials you are using and permissions on the ",
            "workspace or cluster.",
            *(ex.args),
        ]
        raise MsticpyKqlConnectionError(*ex_mssgs, title="authentication failed")

    @staticmethod
    def _raise_unknown_error(ex):
        """Raise an unknown exception."""
        raise MsticpyKqlConnectionError(
            "Another exception was returned by the service",
            *ex.args,
            f"Full exception:\n{ex}",
            title="connection failed",
        )

    def _set_az_auth_option(
        self, mp_az_auth: Union[bool, str, list, None], mp_az_tenant_id: str = None
    ):
        """
        Build connection string with auth elements.

        Parameters
        ----------
        mp_az_auth : Union[bool, str, list, None], optional
            Optional parameter directing KqlMagic to use MSTICPy Azure authentication.
            Values can be:
            - True or "default": use the settings in msticpyconfig.yaml 'Azure' section
            - auth_method: single auth method name ('msi', 'cli', 'env' or 'interactive')
            - auth_methods: list of acceptable auth methods from ('msi', 'cli',
                'env' or 'interactive')
        mp_az_tenant_id: str, optional
            Optional parameter specifying a Tenant ID for use by MSTICPy Azure
            authentication.

        """
        # default to default auth methods
        auth_types = self.az_cloud_config.auth_methods
        # override if user-supplied methods on command line
        if isinstance(mp_az_auth, str) and mp_az_auth != "default":
            auth_types = [mp_az_auth]
        elif isinstance(mp_az_auth, list):
            auth_types = mp_az_auth
        # get current credentials
        creds = az_connect(auth_methods=auth_types, tenant_id=mp_az_tenant_id)
        if only_interactive_cred(creds.modern):
            print("Check your default browser for interactive sign-in prompt.")

        endpoint_uri = self._get_endpoint_uri()
        endpoint_token_uri = f"{endpoint_uri}.default"
        # obtain token for the endpoint
        with contextlib.suppress(ClientAuthenticationError):
            token = creds.modern.get_token(
                endpoint_token_uri, tenant_id=mp_az_tenant_id
            )
            # set the token values in the namespace
            endpoint_token = {
                "access_token": token.token,
                "token_type": "Bearer",
                "resource": endpoint_uri,
            }
            self._set_kql_option("try_token", endpoint_token)

    def _get_endpoint_uri(self):
        return self.az_cloud_config.log_analytics_uri
